
import os, sys
from pathlib import Path
pydir = Path(os.path.abspath(__file__)).parent
import json

import damei as dm

logger = dm.get_logger('ModelHub')

class ModelHub(object):

    def __init__(self) -> None:
        self.urls = f'{pydir.parent}/configs/urls/hub_models.json'
        self._keys = None  # i.e model names
        self._items = None  # i.e model dict

    def _read_json(self):
        with open(self.urls, 'r') as f:
            data = json.load(f)
        self._keys = list(data.keys())
        self._items = data
        return data

    @property
    def keys(self):
        if self._keys is None:
            self._read_json()
        return self._keys

    @property
    def items(self):
        if self._items is None:
            self._read_json()
        return self._items

    def list_remote_models(self, **kwargs):
        """按类别统计"""
        info = {}
        for k, v in self.items.items():
            name = k  # i.e. model name
            category = v['category']  # i.e. model category
            if category not in info.keys():
                info[category] = [name]
                continue
            info[category].append(name)
        return info

    def download(self, *args, **kwargs):
        name = kwargs.get('name', args[0])
        url = self.items[name]['url']
        
        code = f'git clone {url}'
        logger.info(f'Run code: {code}')
        os.system(code)
        




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .parts import (
    Parts,
    AsyncParts,
    PartsWithRawResponse,
    AsyncPartsWithRawResponse,
    PartsWithStreamingResponse,
    AsyncPartsWithStreamingResponse,
)
from .uploads import (
    Uploads,
    AsyncUploads,
    UploadsWithRawResponse,
    AsyncUploadsWithRawResponse,
    UploadsWithStreamingResponse,
    AsyncUploadsWithStreamingResponse,
)

__all__ = [
    "Parts",
    "AsyncParts",
    "PartsWithRawResponse",
    "AsyncPartsWithRawResponse",
    "PartsWithStreamingResponse",
    "AsyncPartsWithStreamingResponse",
    "Uploads",
    "AsyncUploads",
    "UploadsWithRawResponse",
    "AsyncUploadsWithRawResponse",
    "UploadsWithStreamingResponse",
    "AsyncUploadsWithStreamingResponse",
]



"""
模型状态管理器
负责管理模型的启用/禁用状态，并持久化到 YAML 文件
"""
import os
import time
import yaml
import threading
from typing import Dict, List, TYPE_CHECKING
from dataclasses import dataclass, field, asdict

if TYPE_CHECKING:
    from ._worker_class import CommonWorker


@dataclass
class ModelStatusConfig:
    """单个模型的状态配置"""
    model_name: str
    enabled: bool = True
    provider: str = "Uncategorized"
    last_updated: float = field(default_factory=lambda: time.time())


class ModelStatusManager:
    """
    模型状态管理器

    功能：
    - 管理模型的启用/禁用状态
    - 持久化状态到 YAML 文件
    - 按提供者分组模型
    - 线程安全的状态读写
    """

    def __init__(self, worker: 'CommonWorker', config_file: str):
        """
        初始化模型状态管理器

        Args:
            worker: CommonWorker 实例
            config_file: 配置文件路径（model_config.yaml）
        """
        self.worker = worker
        self.config_file = config_file
        self._model_status: Dict[str, bool] = {}  # 内存缓存 {model_name: enabled}
        self._lock = threading.Lock()  # 线程锁

        # 加载或创建配置文件
        self._load_or_create_config()

    def is_model_enabled(self, model_name: str) -> bool:
        """
        检查模型是否启用

        Args:
            model_name: 模型名称

        Returns:
            bool: True 表示启用，False 表示禁用
        """
        with self._lock:
            return self._model_status.get(model_name, True)  # 默认启用

    def set_model_status(self, model_name: str, enabled: bool):
        """
        设置模型状态并持久化到文件

        Args:
            model_name: 模型名称
            enabled: True 启用，False 禁用
        """
        with self._lock:
            old_status = self._model_status.get(model_name)
            self._model_status[model_name] = enabled

            # 记录日志
            if hasattr(self.worker, 'logger'):
                action = "enabled" if enabled else "disabled"
                self.worker.logger.info(
                    f"Model '{model_name}' has been {action} "
                    f"(previous status: {old_status})"
                )

            # 持久化到文件
            self._save_config()

    def get_models_by_provider(self) -> Dict[str, List[Dict]]:
        """
        按提供者分组模型

        Returns:
            Dict[str, List[Dict]]: {
                "hepai": [{"name": "hepai/gpt-4", "enabled": True}, ...],
                "Uncategorized": [{"name": "local-llama", "enabled": False}, ...]
            }
        """
        providers: Dict[str, List[Dict]] = {}

        for model in self.worker.models:
            model_name = model.name
            provider = self._extract_provider(model_name)

            if provider not in providers:
                providers[provider] = []

            providers[provider].append({
                "name": model_name,
                "enabled": self.is_model_enabled(model_name)
            })

        return providers

    def get_all_model_status(self) -> Dict[str, bool]:
        """
        获取所有模型的状态

        Returns:
            Dict[str, bool]: {model_name: enabled}
        """
        with self._lock:
            return self._model_status.copy()

    def _extract_provider(self, model_name: str) -> str:
        """
        从模型名称中提取提供者

        规则：
        - 包含 "/" 的模型名：提取 "/" 前部分（如 "hepai/gpt-4" → "hepai"）
        - 不包含 "/" 的模型名：归类为 "Uncategorized"

        Args:
            model_name: 模型名称

        Returns:
            str: 提供者名称
        """
        if "/" in model_name:
            return model_name.split("/")[0]
        return "Uncategorized"

    def _load_or_create_config(self):
        """加载或创建配置文件"""
        if os.path.exists(self.config_file):
            try:
                self._load_config()
                if hasattr(self.worker, 'logger'):
                    self.worker.logger.info(
                        f"Model status config loaded from: {self.config_file}"
                    )
            except Exception as e:
                # 配置文件损坏，备份并重建
                if hasattr(self.worker, 'logger'):
                    self.worker.logger.error(
                        f"Failed to load config from {self.config_file}: {e}. "
                        f"Creating backup and rebuilding..."
                    )
                self._backup_and_rebuild()
        else:
            # 首次运行，创建配置文件
            self._create_default_config()
            if hasattr(self.worker, 'logger'):
                self.worker.logger.info(
                    f"Model status config created at: {self.config_file}"
                )
            else:
                print(f"Model status config created at: {self.config_file}")

    def _load_config(self):
        """从文件加载配置"""
        with open(self.config_file, 'r', encoding='utf-8') as f:
            config_data = yaml.safe_load(f)

        if not config_data or 'models' not in config_data:
            raise ValueError("Invalid config file format")

        # 加载模型状态
        models_config = config_data.get('models', {})
        for model_name, model_data in models_config.items():
            if isinstance(model_data, dict):
                self._model_status[model_name] = model_data.get('enabled', True)
            else:
                self._model_status[model_name] = True

        # 检查是否有新增的模型（在 worker.models 中但不在配置文件中）
        for model in self.worker.models:
            if model.name not in self._model_status:
                self._model_status[model.name] = True  # 新模型默认启用

    def _create_default_config(self):
        """创建默认配置文件（所有模型默认启用）"""
        for model in self.worker.models:
            self._model_status[model.name] = True

        self._save_config()

    def _save_config(self):
        """保存配置到文件"""
        config_data = {
            'version': '1.0',
            'last_modified': time.time(),
            'models': {}
        }

        for model in self.worker.models:
            model_name = model.name
            config_data['models'][model_name] = {
                'enabled': self._model_status.get(model_name, True),
                'provider': self._extract_provider(model_name),
                'last_updated': time.time()
            }

        # 写入文件（使用临时文件 + 原子替换保证安全性）
        temp_file = self.config_file + '.tmp'
        try:
            with open(temp_file, 'w', encoding='utf-8') as f:
                yaml.dump(config_data, f, allow_unicode=True, default_flow_style=False)

            # 原子替换
            os.replace(temp_file, self.config_file)
        except Exception as e:
            # 清理临时文件
            if os.path.exists(temp_file):
                os.remove(temp_file)
            raise e

    def _backup_and_rebuild(self):
        """备份损坏的配置文件并重建"""
        # 备份
        if os.path.exists(self.config_file):
            backup_file = f"{self.config_file}.backup.{int(time.time())}"
            try:
                os.rename(self.config_file, backup_file)
                if hasattr(self.worker, 'logger'):
                    self.worker.logger.info(f"Backup created: {backup_file}")
            except Exception as e:
                if hasattr(self.worker, 'logger'):
                    self.worker.logger.error(f"Failed to create backup: {e}")

        # 重建
        self._create_default_config()



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List
from typing_extensions import Required, TypedDict

from .run_step_include import RunStepInclude

__all__ = ["StepRetrieveParams"]


class StepRetrieveParams(TypedDict, total=False):
    thread_id: Required[str]

    run_id: Required[str]

    include: List[RunStepInclude]
    """A list of additional fields to include in the response.

    Currently the only supported value is
    `step_details.tool_calls[*].file_search.results[*].content` to fetch the file
    search result content.

    See the
    [file search tool documentation](https://platform.openai.com/docs/assistants/tools/file-search#customizing-file-search-settings)
    for more information.
    """




from hepai.types import LLMRemoteModel, LLMModelConfig
from hepai import HepAI
import os
from pathlib import Path
from dataclasses import asdict
from dotenv import load_dotenv

here = Path(__file__).parent
load_dotenv(f'{here.parent.parent.parent}/.env')


def load_models(model_config: "LLMModelConfig"):
    """加载模型配置文件中的模型"""
    if model_config.config_file is None:
        # 动态获取所有可用模型
        api_key = model_config.api_key
        base_url = model_config.base_url
        
        client = HepAI(api_key=api_key, base_url=base_url)
        try:
            models_res = client.models.list()
            model_names = list(set([x.id.lower() for x in models_res]))
            model_names = sorted(model_names)
            
            models = []

            for model in model_names:
                engine = model
                provider = get_provider_by_model_name(model)
                if provider is None:
                    print(f"Skipping unknown provider model: {model}")
                    continue
                model_name = f"{provider}/{engine}"  # 设置为provider/engine格式
                # model_name = model
                
                # model_name = model
                
                cfg = LLMModelConfig(
                    name=model_name,
                    engine=engine,
                    base_url=model_config.base_url,
                    _api_key=model_config._api_key,
                    proxy=model_config.proxy,
                    need_external_api_key=model_config.need_external_api_key,
                    permission=model_config.permission,
                    version=model_config.version,
                    enable_async=model_config.enable_async,
                    test=model_config.test
                )
                models.append(LLMRemoteModel(config=cfg))
            
            print(f"Successfully loaded {len(models)} models from {base_url}.")
            return models
            
        except Exception as e:
            print(f"Failed to load models from API: {e}")
            # 如果API调用失败，回退到原来的单模型模式
            models = [LLMRemoteModel(config=model_config)]
    else:
        import yaml
        with open(model_config.config_file, "r", encoding="utf-8") as f:
            config_dict = yaml.safe_load(f)
        model_list = config_dict.get("models_list", [])
        assert len(model_list) > 0, "No models found in the configuration file"
        # 假设yaml顶层是一个列表，每个元素是一个模型配置
        models = []
        for m_cfg in model_list:
            # 只取LLMModelConfig支持的字段
            # 字段映射：model_name -> name
            if "model_name" in m_cfg:
                m_cfg["name"] = m_cfg.pop("model_name")
            allowed_keys = LLMModelConfig.__dataclass_fields__.keys()
            filtered_cfg = {k: v for k, v in m_cfg.items() if k in allowed_keys}
            exist_cfg_dict = asdict(model_config)
            exist_cfg_dict.update(filtered_cfg)  # 使用yaml中的字段覆盖默认配置
            cfg = LLMModelConfig(**exist_cfg_dict)
            models.append(LLMRemoteModel(config=cfg))
    return models

def get_provider_by_model_name(model_name: str) -> str:
    """
    根据模型名字获得provider。
    例如: chatgpt类模型->openai, deepseek类模型->deepseek-ai, qwen类模型->aliyun。
    未知模型抛出异常。
    """
    name = model_name.lower()
    
    if "deepseek" in name:
        return "deepseek-ai"
    elif "grok" in name:
        return "xAI"
    elif "ernie" in name or name == 'x1':
        return "baidu"
    elif "360zhinao" in name:
        return "360"
    elif "claude" in name:
        return "anthropic"
    elif "hunyuan" in name:
        return "tencent"
    elif "gemini" in name:
        return "google"
    elif "glm" in name:
        return "zhipu"
    elif "baichuan" in name:
        return "baichuan"
    elif "doubao" in name:
        return "bytedance"
    elif any(x in name for x in ["qwen", "qianwen", "aliyun", "qwq"]):
        return "aliyun"
    elif any(x in name for x in ["minimax", "abab"]):
        return "minimax"
    elif any(x in name for x in ["kimi", "moonshot"]):
        return "moonshot"
    elif "llama" in name:
        return "meta"
    elif "yi" in name:
        return "01-ai"
    # elif any(x in name for x in ["generalv3", "4.0ultra"]):
    elif name in ["generalv3", "4.0ultra", "pro-128k", "max-32k", "lite", 'generalv3.5']:
        return "iflytech"
    elif any(x in name for x in ["gpt", "chatgpt", "openai", "o1", "o3", "o4", "tts-1",
                                 "text-embedding", "dall-e", "whisper-1", "codex-mini", "computer-use-preview"]):
        return "openai"
    elif name in ['max-32k', "lite"]:
        return None
    else:
        return None
        # raise ValueError(f"未知模型provider: {model_name}")


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from ..._models import BaseModel

__all__ = ["FineTuningJobWandbIntegration"]


class FineTuningJobWandbIntegration(BaseModel):
    project: str
    """The name of the project that the new run will be created under."""

    entity: Optional[str] = None
    """The entity to use for the run.

    This allows you to set the team or username of the WandB user that you would
    like associated with the run. If not set, the default entity for the registered
    WandB API key is used.
    """

    name: Optional[str] = None
    """A display name to set for the run.

    If not set, we will use the Job ID as the name.
    """

    tags: Optional[List[str]] = None
    """A list of tags to be attached to the newly created run.

    These tags are passed through directly to WandB. Some default tags are generated
    by OpenAI: "openai/finetune", "openai/{base-model}", "openai/{ftjob-abcdef}".
    """



import os, sys
from pathlib import Path
from easydict import EasyDict

import functools
import collections
import copy
import re
import damei as dm
import hai
import json
import argparse
import importlib
import importlib.util

pydir = Path(os.path.abspath(__file__)).parent
logger = dm.get_logger(__name__)


class PyConfigLoader(collections.UserDict):
# class PyConfigLoader(object):
    """
    从字典、.py文件等获取的配置对象
    v1.0.1 支持元数据
    """

    def __init__(self, cfg_file=None, name=None, root_path=None, cfg_dict=None, **kwargs):
        super(PyConfigLoader, self).__init__()
        # super().__init__()
        work_dir = os.getcwd()
        self.root_path = root_path if root_path else f'{work_dir}'  # 默认当前工作目录
        self._name = name if name else f'{cfg_file}'  # 其实是path

        self._items = dict()
        self._load_items(cfg_file, cfg_dict, **kwargs)  # 从配置文件或配置字典初始化配置，配置添加到_items和对应的属性里
        self.check_items()  # TODO

    def _load_items(self, cfg_file=None, cfg_dict=None, **kwargs):
        if cfg_file is None and cfg_dict is None:
            return

        assert cfg_file or cfg_dict, 'Cfg_file or cfg_dict must be not None'
        assert not (cfg_file and cfg_dict), 'Only one of cfg_file and cfg_dict can have value'
        if cfg_file:
            assert os.path.exists(cfg_file), f'Config file {cfg_file} does not exist'
            self.init_config_by_file(cfg_file)  # 根据配置文件把内容和属性注册到items里
        else:
            self.init_config_by_dict(cfg_dict, **kwargs)

    def init_config_by_dict(self, cfg_file, **kwargs):
        """从字典初始化配置"""
        need_replace = kwargs.get('need_replace', False)  # 是否需要替换 - to _，默认不需要, needed when read config from dm.argparse shi xuyao
        for k, v in cfg_file.items():
            attr, attr_value = k, v
            if need_replace and '-' in attr:
                attr = attr.replace('-', '_')

            self._items[attr] = attr_value
            if isinstance(attr_value, dict):
                setattr(self, attr, EasyDict(self._items[attr]))
            else:
                setattr(self, attr, self._items[attr])

    def __len__(self):
        return len(self.items)

    # def __dict__(self):
    #   return self._items

    def __getitem__(self, item):
        return self.items[item]

    def __repr__(self):
        info = self.info(need_head=False, need_tail=False)
        if info == 'Empty config':
            format_str = f"Empty <class 'hai.Config'> "
        else:
            format_str = f"<class 'hai.Config'> " \
                        f"Configs:\n" \
                        f"{self.info(need_head=False, need_tail=False)}"

        return format_str

    def __missing__(self, key):
        if isinstance(key, str):
            raise KeyError(key)
        return self[str(key)]

    def __contains__(self, key):
        return str(key) in self.data

    def __setitem__(self, key, item):
        # print(f'__setitem__:{key}')
        # self.data[str(key)] = item
        self._items[str(key)] = item
        self.__setattr__(str(key), item)

    @property
    def path(self):
        return self._name

    @property
    def items(self):
        """
        debug: 
            when call self.some_attr = new_value, the new_value will be set to self.__dict__ only,
            but not to self._items.
            To fix this, we need to copy self.__dict__ to self._items every if the value is changed.
        """
        # for k, v in self._items.items():
        #     __dict__v = self.__dict__[k]
        #     if __dict__v != v:
        #         self._items[k] = __dict__v

        # return self._items
        return self.__dict__

    @staticmethod
    def from_anything(*args, **kwargs):
        """
        从字典、.py文件等获取的配置对象
        """
        if len(args) == 1:
            if isinstance(args[0], dict):
                return PyConfigLoader(cfg_dict=args[0], **kwargs)
            elif isinstance(args[0], str):
                return PyConfigLoader(cfg_file=args[0], **kwargs)
            else:
                raise NotImplementedError('Only support dict and str')
        return PyConfigLoader(**kwargs)

    @staticmethod
    def from_file(cfg_file, name=None, root_path=None):
        return PyConfigLoader(cfg_file, name, root_path)

    @staticmethod
    def fromfile(cfg_file, name=None, root_path=None):
        return PyConfigLoader(cfg_file, name, root_path)

    @staticmethod
    def from_dict(cfg_dict, name=None, root_path=None):
        """从字典创建配置对象"""
        name = name if name else 'cfg_from_dict'
        return PyConfigLoader(name=name, root_path=root_path, cfg_dict=cfg_dict)

    @staticmethod
    def from_argparser(args, name=None, root_path=None):
        """从hai.parser.ArgumentParser对象创建配置对象"""
        name = name if name else 'cfg_from_argparser'
        return PyConfigLoader(name=name, root_path=root_path, cfg_dict=args._args_dict)

    @staticmethod
    def from_argparse(opt, name=None, root_path=None):
        name = name if name else 'cfg_from_argparse'
        __dict__ = opt.__dict__
        # print(opt)
        return PyConfigLoader(name=name, root_path=root_path, cfg_dict=__dict__)

    @staticmethod
    def from_dm_argparse(opt, name=None, root_path=None):
        name = name if name else 'cfg_from_dm_argparse'
        __dict__ = opt._args_dict
        return PyConfigLoader(name=name, root_path=root_path, 
                            cfg_dict=__dict__, need_replace=True)


    def info(self, need_head=True, need_tail=True):
        """
        查看当前所有配置信息
        :param need_head: 是否需要包含头部信息，即：Configs:\n
        :param need_tail: 是否需要包含尾部信息，即：\n回车符
        """
        # info_str = f'{self.__repr__()}\n'
        info_dict = self._items
        if info_dict == {}:
            info_str = 'Empty config\n'
        else:
            info_str = f'Configs:\n' if need_head else ''
            info_str += dm.misc.dict2info(info_dict)
        if not need_tail:
            info_str = info_str[:-1]
        return info_str

    def init_config_by_file(self, cfg_file):
        if cfg_file is None:
            return
        rp = self.root_path
        cp = Path(os.path.abspath(cfg_file))  # config path

        if not os.path.exists(cfg_file):
            raise FileNotFoundError(f'{cfg_file} not found')

        # 提取模块路径：如：modules.detection.seyolov5.config
        method = 'method3'
        current_wd = os.getcwd()  # 当前工作目录
        if method == 'method1':
            module_dir = f'{str(cp.parent).replace(f"{rp}/", "").replace("/", ".")}'
            # print(f'rp: {rp} \ncp: {cp}\nm dir: {module_dir}')
            # code = f'from {module_dir} import {cp.stem}'
            code = f'import {module_dir}.{cp.stem} as {cp.stem}'
            # 不能用这种方法，原因是：from modules.xxx.config imort config 过程中会调起modules下的__init__.py,从而调起注册函数，导致重复注册
        elif method == 'method2':  # 方法2：添加临时环境路径
            # 这个方法在工作路径也存在config文件夹时，会导致from config import xxx 导入失败
            sys.path.append(f'{cp.parent.parent}')  # 添加临时环境路径, 如：/home/user/xx/modules/xx
            module_dir = f'{str(cp.parent.stem)}'  # configs文件夹路径，如：configs
            code = f'from {module_dir} import {cp.stem}'
        elif method == 'method3':  # 方法3：添加临时环境路径
            # sys.path.append(f'{cp.parent}')
            sys.path.insert(0, f'{cp.parent}')  # 添加临时环境路径, 如：/home/user/xx/modules/xx
            module_dir = '.'
            os.chdir(f'{cp.parent}')  # 切换工作目录
            # code = f'from {module_dir} import {cp.stem}'
            code = f'import {cp.stem}'
        else:
            raise ValueError(f'invalid method: {method}')

        # print(cfg_file, code)
        """
        exec(code)
        """
        try:
            exec(code)
        except Exception as e:
            msg = f'Error in "exec(code)": {e}. Error variables:\n' \
                  f'    root path: {rp} \n' \
                  f'    config file: {cfg_file} \n' \
                  f'    code: {code} \n' \
                  f'    module dir: {module_dir} \n' \
                  f'    sys.path[0]: {sys.path[0]} ' \
                  f'    sys.path[-1]: {sys.path[-1]}'
            # logger.error(msg)
            raise Exception(msg)
            # sys.exit(1)

        if method == 'method2':
            sys.path.remove(f'{cp.parent.parent}')  # 删除临时环境路径
        elif method == 'method3':
            sys.path.remove(f'{cp.parent}')
            os.chdir(current_wd)  # 切换回工作目录

        # cfg = eval(cp.stem)  # 模块名
        # fix bug in 20251231, adapted to python 3.13+
        spec = importlib.util.spec_from_file_location(cp.stem, str(cp))
        mod = importlib.util.module_from_spec(spec)
        sys.modules[cp.stem] = mod
        spec.loader.exec_module(mod)

        cfg = mod

        # 读取文件获取顶格写并且不以#开头的属性
        with open(cfg_file, 'r', encoding='utf-8') as f:
            data = f.readlines()
        data = [x for x in data if not (x.startswith(' ') or x.startswith('#'))]  # 非顶格的不要，# 开头的不要
        data = '\n'+'\n'.join(data)
        attrs = re.findall(pattern=r"\s[A-Za-z]\w+ ?= ?", string=data)
        # The egular expression: \s[A-Za-z]\w+ ?= ? to match: ' attr = '
        # i.e. start with a space, then a letter (a to z or A to Z), 
        # then any number of letters, numbers or underscores, 
        # then a space, then an equal sign, then a space
        attrs = [x.split('=')[0].replace('\n', '').strip() for x in attrs]

        # 注册属性到_items里，注册属性到self的属性里。
        for attr in attrs:
            # print('xxx', attr)
            if hasattr(cfg, attr):
                exec(f"self._items[attr] = cfg.{attr}")
                attr_value = self._items[attr]
                if isinstance(attr_value, dict):
                    setattr(self, attr, EasyDict(self._items[attr]))
                else:
                    setattr(self, attr, attr_value)
            else:
                pass

    def check_items(self):
        """初始化末，检查所有配置项，处理：缩略语~转为真实路径。
        递归方法，如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环
        """
        # print(self._items)
        for k, v in self._items.items():
            new_v = self.recursive_func(v)
            self._items[k] = new_v
            # 更新属性
            if hasattr(self, k):
                delattr(self, k)
            setattr(self, k, EasyDict(new_v)) if isinstance(new_v, dict) else setattr(self, k, new_v)

    def recursive_func(self, value):
        """递归函数,如果v类型是str,判断替换，如果循环完了，结束，如果v的类型是dict，继续循环"""
        if value is None:
            return None
        elif isinstance(value, str):
            return value.replace('~', str(Path.home()))
        elif isinstance(value, int):
            return value
        elif isinstance(value, float):
            return value
        elif isinstance(value, bool):
            return value
        elif isinstance(value, list):
            new_list = [self.recursive_func(x) for x in value]
            return new_list
        elif isinstance(value, dict):
            new_dict = {}
            for k, v in value.items():
                new_dict[k] = self.recursive_func(v)
            return new_dict
        # elif isinstance(value, argparse.Namespace):  # TODO: suport argparse.Namespace
            # items = value.items()
        elif isinstance(value, hai.Config):
            new_dict = {}
            for k, v in value._items.items():
                new_dict[k] = self.recursive_func(v)
            new_config = PyConfigLoader.from_dict(new_dict)
            return new_config
        else:
            raise TypeError(f'Config file only support basic python types but the value type is: {type(value)}.')

    def update_item(self, attrs, value):
        """更新配置条目
        attrs: 长度为0时代表，更新全部value，长度为1时，代表第1个元素是第一层，
        """
        if len(attrs) == 0:
            self.merge(value)
        elif len(attrs) == 1:
            self._items[attrs[0]] = value
        elif len(attrs) == 2:
            self._items[attrs[0]][attrs[1]] = value
        elif len(attrs) == 3:
            self._items[attrs[0]][attrs[1]][attrs[2]] = value
        else:
            self._items[attrs[0]][attrs[1]][attrs[2]][attrs[3]] = value
        self.check_items()

    def merge2opt(self, opt):
        """
        合并配置文件到opt, 合并规则：没想好，先放着
        return: opt
        """
        assert isinstance(opt, argparse.Namespace), 'opt must be argparse.Namespace'
        raise NotImplementedError(f'未实现')

    def merge(self, cfg2, **kwargs):
        """合并另一个配置文件到内部，是inplace的"""
        if cfg2 is None:
            return self

        if isinstance(cfg2, dict):
            cfg2 = PyConfigLoader.from_dict(cfg2)

        for k2, v2 in cfg2._items.items():
            merged_v = self.recursive_func2(v2, k2=k2, sub_items=self._items)
            # self._items[k2] = merged_v
            self.__dict__[k2] = merged_v
        self.check_items()
        return self

    def recursive_func2(self, v2, k2=None, sub_items=None):
        """v1是之前的值，v2是现在的值"""
        if k2 and (k2 not in list(sub_items.keys())):
            return v2
        else:
            v1 = sub_items[k2]

            if type(v1) != type(v2):
                return v2
            f = functools.partial(isinstance, v1)
            if f(str) or f(int) or f(float) or f(bool):
                return v2
            elif f(list):
                tmp = [x for x in v2 if x not in v1]
                return v1 + tmp
            elif f(dict):  # v1v2都是dict
                new_dict = v1  # 保留v1
                for kk, vv in v2.items():
                    new_dict[kk] = self.recursive_func2(vv, k2=kk, sub_items=v1)
                return new_dict

    def map_keys(self, map_dict, new=True):
        """map_dict: {old_key: new_key}"""
        assert isinstance(map_dict, dict), 'map_dict must be dict'
        if new:
            new_cfg = copy.deepcopy(self)
            new_cfg.map_keys(map_dict, new=False)
            return new_cfg
        else:
            for k, v in map_dict.items():  # k是旧的，v是新的
                if k in self._items:
                    self.__dict__[v] = self.items[k]
                    del self.__dict__[k]
        

    def keys(self):
        return self._items.keys()

    def get(self, key, default=None):
        if key in list(self._items.keys()):
            return self._items[key]
        else:
            return default

    def to_dict(self):
        # raise NotImplementedError('to_dict is not implemented.')
        return self._items

    def to_json(self):
        return json.dumps(self._items, indent=4, ensure_ascii=False)



"""
模型管理路由组
提供模型启用/禁用的 API 端点
"""
import os
import time
from typing import List, Callable, Dict
from dataclasses import field, dataclass
from fastapi import APIRouter, Depends, HTTPException
from fastapi.requests import Request
from fastapi.responses import HTMLResponse

from ..worker_app import HWorkerAPP
from ..utils import read_request_body
from ..singletons import authorizer

# 管理员认证依赖
admin_auth: Callable = Depends(authorizer.admin_auth)


@dataclass
class ModelManagerRouterGroup:
    """模型管理路由组"""
    name: str = "model_manager"
    prefix: str = "/apiv2"
    tags: List[str] = field(default_factory=lambda: ["model_manager"])
    router: APIRouter = field(default_factory=APIRouter)
    parent_app: HWorkerAPP = None  # type: ignore

    def __post_init__(self):
        """初始化路由"""
        rt = self.router

        # Dashboard 页面（无需认证，认证在前端完成）
        rt.get("/dashboard")(self.dashboard_page)

        # API 端点（需要管理员认证）
        rt.get("/models/grouped", dependencies=[admin_auth])(self.list_models_grouped)
        rt.post("/models/{model_name}/enable", dependencies=[admin_auth])(self.enable_model)
        rt.post("/models/{model_name}/disable", dependencies=[admin_auth])(self.disable_model)
        rt.post("/models/batch_update", dependencies=[admin_auth])(self.batch_update_models)
        rt.get("/models/status", dependencies=[admin_auth])(self.get_all_status)

    async def dashboard_page(self):
        """返回 Dashboard HTML 页面"""
        html_file = os.path.join(os.path.dirname(__file__), "../html", "dashboard.html")
        try:
            with open(html_file, 'r', encoding='utf-8') as f:
                content = f.read()
            return HTMLResponse(content=content)
        except Exception as e:
            raise HTTPException(
                status_code=500,
                detail=f"Failed to load dashboard page: {e}"
            )

    async def list_models_grouped(self):
        """
        按提供者分组返回所有模型及状态

        Returns:
            {
                "success": True,
                "data": {
                    "hepai": [{"name": "hepai/gpt-4", "enabled": True}, ...],
                    "Uncategorized": [{"name": "local-llama", "enabled": False}, ...]
                },
                "timestamp": 1703001234.567
            }
        """
        manager = self.parent_app.worker.model_status_manager
        grouped = manager.get_models_by_provider()

        return {
            "success": True,
            "data": grouped,
            "timestamp": time.time()
        }

    async def enable_model(self, model_name: str):
        """
        启用指定模型

        Args:
            model_name: 模型名称

        Returns:
            {
                "success": True,
                "message": "Model xxx enabled",
                "model_name": "xxx",
                "enabled": True
            }
        """
        manager = self.parent_app.worker.model_status_manager

        # 检查模型是否存在
        if model_name not in self.parent_app.worker._model_map:
            raise HTTPException(
                status_code=404,
                detail=f"Model '{model_name}' not found"
            )

        # 启用模型
        manager.set_model_status(model_name, enabled=True)

        return {
            "success": True,
            "message": f"Model '{model_name}' enabled",
            "model_name": model_name,
            "enabled": True
        }

    async def disable_model(self, model_name: str):
        """
        禁用指定模型

        注意：正在处理的请求会完成，排队中的请求会在获取信号量后被拒绝

        Args:
            model_name: 模型名称

        Returns:
            {
                "success": True,
                "message": "Model xxx disabled",
                "model_name": "xxx",
                "enabled": False,
                "note": "..."
            }
        """
        manager = self.parent_app.worker.model_status_manager

        # 检查模型是否存在
        if model_name not in self.parent_app.worker._model_map:
            raise HTTPException(
                status_code=404,
                detail=f"Model '{model_name}' not found"
            )

        # 禁用模型
        manager.set_model_status(model_name, enabled=False)

        return {
            "success": True,
            "message": f"Model '{model_name}' disabled",
            "model_name": model_name,
            "enabled": False,
            "note": "In-progress requests will complete, queued requests will be rejected"
        }

    async def batch_update_models(self, request: Request):
        """
        批量更新多个模型的状态

        Request Body:
            {
                "updates": [
                    {"model_name": "xxx", "enabled": true},
                    {"model_name": "yyy", "enabled": false},
                    ...
                ]
            }

        Returns:
            {
                "success": True,
                "results": [
                    {"model_name": "xxx", "enabled": true, "success": True},
                    {"model_name": "yyy", "enabled": false, "success": True}
                ],
                "count": 2
            }
        """
        body = await read_request_body(request)
        updates = body.get("updates", [])

        if not updates:
            raise HTTPException(
                status_code=400,
                detail="No updates provided"
            )

        manager = self.parent_app.worker.model_status_manager
        results = []

        for update in updates:
            model_name = update.get("model_name")
            enabled = update.get("enabled")

            if not model_name:
                results.append({
                    "model_name": model_name,
                    "success": False,
                    "error": "model_name is required"
                })
                continue

            if enabled is None:
                results.append({
                    "model_name": model_name,
                    "success": False,
                    "error": "enabled is required"
                })
                continue

            # 检查模型是否存在
            if model_name not in self.parent_app.worker._model_map:
                results.append({
                    "model_name": model_name,
                    "success": False,
                    "error": f"Model '{model_name}' not found"
                })
                continue

            # 更新状态
            try:
                manager.set_model_status(model_name, enabled)
                results.append({
                    "model_name": model_name,
                    "enabled": enabled,
                    "success": True
                })
            except Exception as e:
                results.append({
                    "model_name": model_name,
                    "success": False,
                    "error": str(e)
                })

        # 如果 worker 已注册到 controller，发送一次心跳更新状态
        worker = self.parent_app.worker
        if not worker.config.no_register:
            try:
                # 在后台线程中发送心跳，避免阻塞响应
                import asyncio
                loop = asyncio.get_event_loop()
                loop.run_in_executor(None, worker.register_to_controller, True)

                if hasattr(worker, 'logger'):
                    worker.logger.info("Model status updated, heartbeat sent to controller")
            except Exception as e:
                if hasattr(worker, 'logger'):
                    worker.logger.warning(f"Failed to send heartbeat after model status update: {e}")

        return {
            "success": True,
            "results": results,
            "count": len(results)
        }

    async def get_all_status(self):
        """
        获取所有模型的状态（用于验证密码）

        Returns:
            {
                "success": True,
                "models": {
                    "hepai/gpt-4": True,
                    "local-llama": False,
                    ...
                },
                "timestamp": 1703001234.567
            }
        """
        manager = self.parent_app.worker.model_status_manager
        all_status = manager.get_all_model_status()

        return {
            "success": True,
            "models": all_status,
            "timestamp": time.time()
        }



from typing import Dict, List, Literal, Optional, Union, Generator, Callable
from dataclasses import dataclass, field
import os

from fastapi import FastAPI, APIRouter, Request, HTTPException
from fastapi import UploadFile, File, Form, Query, Body
from fastapi.background import BackgroundTasks
from fastapi.responses import JSONResponse, StreamingResponse, FileResponse, HTMLResponse

import asyncio
from asyncio import Semaphore
import threading
from pydantic import BaseModel
from logging import Logger
import httpx
import inspect
import time
from concurrent.futures import ThreadPoolExecutor
# import markdown

from ._worker_class import HWorkerConfig, HRemoteModel, CommonWorker, ModelResourceInfo
from ._related_class import WorkerStoppedInfo, WorkerInfoRequest
from . import utils
from .singletons import authorizer

#
from .mcp_adapter.utils import build_mcp_kwargs_for_starlette

class FunctionParamsItem(BaseModel):
    args: List = []
    kwargs: Dict = {}


def get_fastapi_init_params():
    # 获取 FastAPI 的 __init__ 方法签名
    init_signature = inspect.signature(FastAPI.__init__)
    # 提取参数名（排除 'self' 和可变参数）
    params = []
    for name, param in init_signature.parameters.items():
        if name == "self":
            continue
        if param.kind in (
            inspect.Parameter.VAR_POSITIONAL,  # *args
            inspect.Parameter.VAR_KEYWORD       # **kwargs
        ):
            continue
        params.append(name)
    return params

class HWorkerAPP(FastAPI):
    """
    FastAPI app for worker
    多线程运行，支持多个模型
    """

    def __init__(
            self,
            models: HRemoteModel | List[HRemoteModel], 
            worker_config: HWorkerConfig = None,  # Alias of config
            logger: Logger = None,
            **worker_overrides,
            ):
        # 获取父类的所有可能参数并传递给父类
        # 获取 FastAPI 接受的参数名
        fastapi_params = get_fastapi_init_params()
        # 筛选出有效参数
        fastapi_kwargs = {
            k: v for k, v in worker_overrides.items()
            if k in fastapi_params
        }
        # 从 worker_overrides 中删除 FastAPI 的参数
        for k in fastapi_kwargs.keys():
            worker_overrides.pop(k)

        # 如果有任意model中enable_mcp为True，则启用MCP路由
        mcp_kwargs = build_mcp_kwargs_for_starlette(models, route_prefix=worker_config.route_prefix)
        fastapi_kwargs.update(mcp_kwargs)
        
        super().__init__(**fastapi_kwargs)
        
        
        self.logger = self.get_logger(logger)
        worker_config = worker_config if worker_config is not None else HWorkerConfig()
        assert isinstance(worker_config, HWorkerConfig), f"worker_config should be an instance of HWorkerConfig"
        worker_config.update_from_dict(worker_overrides)
        
        # 用于控制模型访问的信号量 - 改为每个模型独立的信号量
        self.limit_model_concurrency = worker_config.limit_model_concurrency
        self.model_semaphores: Dict[str, Semaphore] = {}  # 每个模型独立的信号量
        self.global_counter = 0
        self._model_lookup_cache: Dict[str, int] = {}  # 模型查找缓存
        
        # 初始化线程池用于I/O密集型操作
        self._thread_pool = ThreadPoolExecutor(max_workers=100, thread_name_prefix="worker_io")
        if worker_config.enable_secret_key:
            seed = utils.get_simple_machine_seed()
            worker_secret_key = utils.gen_one_key(prefix='sk-', lenth=47, seed=seed)
            self.logger.info(f"Worker secret key: `{worker_secret_key}`, please pass it in the `Authorization` header when you call this worker.")
            authorizer.secret_key = worker_secret_key
        else:
            worker_secret_key = None
        self.worker_secret_key = worker_secret_key
        # 初始化模型信号量和缓存
        self._init_model_resources(models=models)
        self._init_routers(config=worker_config)

        self.worker = CommonWorker(
            app=self, models=models, worker_config=worker_config,
            logger=self.logger)

        # 生成管理员密码（在 worker 创建之后，因为需要 worker_id）
        if worker_config.enable_secret_key:
            admin_seed = utils.get_simple_machine_seed(extra_indicators=[self.worker.worker_id])
            self.admin_password = utils.gen_one_key(prefix='admin-', lenth=20, seed=admin_seed)
            self.logger.info(f"Admin password for model management: `{self.admin_password}`")
            authorizer.admin_password = self.admin_password
        else:
            self.admin_password = None


        self._docs_content_cache = None  # 缓存docs内容
        self._html_mk_content_cache = None  # 缓存markdown内容
        self._cache_lock = asyncio.Lock()  # 缓存锁防止竞态条件

    def _init_model_resources(self, models: Union[HRemoteModel, List[HRemoteModel]]):
        """初始化模型资源：信号量和查找缓存"""
        if not isinstance(models, List):
            models = [models]
        for i, model in enumerate(models):
            model_name = model.name
            # 为每个模型创建独立的信号量
            self.model_semaphores[model_name] = asyncio.Semaphore(self.limit_model_concurrency)
            # 建立模型名到索引的快速查找缓存
            self._model_lookup_cache[model_name] = i
        
    
    def _init_routers(self, config: HWorkerConfig):
        
        # 1 index router
        index_router = APIRouter(prefix="", tags=["base"])
        index_router.get("/")(self.index)
        self.include_router(index_router)

        # 2 worker router
        worker_router = self.get_worker_router(router_prefix=config.route_prefix)
        self.include_router(worker_router)
        
        # 3 llm router
        if config.enable_llm_router:
            from .routers.llm_router import LLMRouterGroup
            llm_rg = LLMRouterGroup(prefix=config.route_prefix, parent_app=self)
            self.include_router(llm_rg.router, prefix=llm_rg.prefix, tags=llm_rg.tags)

            from .routers.anthropic_router import AnthropicRouterGroup
            anthropic_rg = AnthropicRouterGroup(prefix=config.route_prefix, parent_app=self)
            self.include_router(anthropic_rg.router, prefix=anthropic_rg.prefix, tags=anthropic_rg.tags)

        # 4 model manager router
        from .routers.model_manager_router import ModelManagerRouterGroup
        model_manager_rg = ModelManagerRouterGroup(prefix=config.route_prefix, parent_app=self)
        self.include_router(model_manager_rg.router, prefix=model_manager_rg.prefix, tags=model_manager_rg.tags)

        # 5 mcp router
        # if config.enable_mcp:
        #     from .mcp_adapter.mcp_router import MCPRouterGroup
        #     mcp_rg = MCPRouterGroup(prefix=config.route_prefix, parent_app=self)
        #     self.include_router(mcp_rg.router, prefix=mcp_rg.prefix, tags=mcp_rg.tags)
        
    def get_worker_router(self, router_prefix: str = ""):
        # router_prefix = self.worker.config_dict.get("route_prefix", "/apiv2")
        router = APIRouter(prefix=router_prefix, tags=["worker"])
        router.post("/worker_unified_gate/")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{function}")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{model}/{function}")(self.worker_unified_gate)  # 多模型模式下，需要指定模型
        router.post("/worker/unified_gate/")(self.worker_unified_gate)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.post("/worker/unified_gate/{function}")(self.worker_unified_gate)
        router.post("/worker/unified_gate/{model}/{function}")(self.worker_unified_gate)  # 多模型模式下，需要指定模型

        router.post("/worker/get_worker_info")(self.get_worker_info)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.get("/worker/models")(self.get_models)
        router.get("/worker/monitor_status")(self.get_monitor_status)  # 监控状态接口
        router.get("/worker/get_status")(self.worker_get_status)
        router.get("/worker_get_status")(self.worker_get_status)
        router.post("/worker/shutdown")(self.shutdown_worker)
        router.post("/shutdown_worker")(self.shutdown_worker)
        
        return router

    @classmethod
    def get_logger(cls, logger: Logger = None):
        if logger is None:
            try:
                from ...utils._logger import Logger
                logger = Logger.get_logger("worker_app.py")
            except:
                import logging
                logger = logging.getLogger("worker_app.py")
        return logger

    @property
    def host(self):
        return self.worker.network_info.host
    
    @property
    def port(self):
        return self.worker.network_info.port
    

    def get_queue_length(self, model_name: str = None):
        """获取队列长度，如果指定模型则返回该模型的队列长度，否则返回总队列长度"""
        if model_name:
            model_semaphore = self.model_semaphores.get(model_name)
            if model_semaphore is None:
                return 0
            try:
                _value = model_semaphore._value
                _num_waiters = len(model_semaphore._waiters) if hasattr(model_semaphore, "_waiters") and model_semaphore._waiters is not None else 0
                return self.limit_model_concurrency - _value + _num_waiters
            except Exception:
                return 0
        else:
            # 返回所有模型的总队列长度
            total_queue = 0
            for semaphore in self.model_semaphores.values():
                try:
                    _value = semaphore._value
                    _num_waiters = len(semaphore._waiters) if hasattr(semaphore, "_waiters") and semaphore._waiters is not None else 0
                    total_queue += self.limit_model_concurrency - _value + _num_waiters
                except Exception:
                    continue
            return total_queue
            
    def release_model_semaphore(self, model_semaphore: Semaphore):
        """释放指定模型的信号量"""
        if model_semaphore is not None:
            model_semaphore.release()
            
    async def index(self):
        # 返回监控面板HTML页面
        try:
            html_file_path = os.path.join(os.path.dirname(__file__), "html", "worker_index.html")
            with open(html_file_path, 'r', encoding='utf-8') as f:
                content = f.read()
            return HTMLResponse(content=content)
        except Exception as e:
            self.logger.error(f"Failed to load HTML file: {e}")
            # 如果HTML文件加载失败，返回简单的错误页面
            error_content = f"""
            <html>
                <head>
                    <title>HepAI Worker - Error</title>
                </head>
                <body>
                    <h1>Worker Index Page Error</h1>
                    <p>无法加载监控页面: {str(e)}</p>
                    <p>Worker ID: {self.worker.worker_id}</p>
                    <p>API Documentation: <a href="/docs">/docs</a></p>
                </body>
            </html>
            """
            return HTMLResponse(content=error_content)

    async def worker_unified_gate(
            self,
            function_params: FunctionParamsItem,
            model: str = None,
            function: str = "__call__",
            ):
        self.global_counter += 1

        # 确定使用的模型
        if model is None:
            if len(self.worker.models) == 1:
                model = self.worker.models[0].name
            else:
                raise HTTPException(status_code=400, detail="Model parameter is required when multiple models available")

        # 检查模型是否存在
        if model not in self.model_semaphores:
            raise HTTPException(status_code=503, detail=f"Model '{model}' not found")

        # 【第一次检查】检查模型是否被禁用（获取信号量前）
        if not self.worker.model_status_manager.is_model_enabled(model):
            raise HTTPException(
                status_code=403,
                detail=f"Model '{model}' is currently disabled by administrator"
            )

        # 获取该模型的信号量
        model_semaphore = self.model_semaphores[model]
        await model_semaphore.acquire()

        # 【第二次检查】再次检查模型状态（防止在等待期间被禁用）
        if not self.worker.model_status_manager.is_model_enabled(model):
            self.release_model_semaphore(model_semaphore)
            raise HTTPException(
                status_code=403,
                detail=f"Model '{model}' was disabled while request was queued"
            )

        # print(f"[{self.global_counter}] Acquired semaphore for model '{model}'. Current queue length: {self.get_queue_length(model)}")

        try:
            rst = await self.worker.unified_gate_async(
                model=model,
                function=function,
                args=function_params.args,
                kwargs=function_params.kwargs)
        except Exception as e:
            # await self.release_model_semaphore(model)
            # self.release_model_semaphore(model)
            self.release_model_semaphore(model_semaphore)
            raise e

        # background_tasks.add_task(self.release_model_semaphore, model_semaphore)
        self.release_model_semaphore(model_semaphore)
        return rst
    
    async def worker_get_status(self):
        """获取worker的状态信息，即WorkerStatusInfo"""
        return self.worker.get_status_info().to_dict()
    
    async def get_worker_info(
            self, 
            worker_info_request: WorkerInfoRequest,
            # user_auth: HAPIKeyAuth = api_key_auth,
            ) -> JSONResponse:
        """
        与controller的worker_info接口一致，以便client调用
        """
        # return self.worker.get_worker_info()
        rst = self.worker.get_worker_info()
        return rst

    async def get_models(self):
        mrs: List[ModelResourceInfo] = self.worker.get_model_resource_info()
        
        return {
            "object": "list",
            "data": [mr.to_dict() for mr in mrs],
        }
    
    async def get_monitor_status(self):
        """获取实时监控状态信息"""
        try:
            # 获取每个模型的并发情况
            model_status = {}
            total_queue = 0
            total_active = 0
            
            for model_name, semaphore in self.model_semaphores.items():
                try:
                    available = semaphore._value
                    waiters = len(semaphore._waiters) if hasattr(semaphore, "_waiters") and semaphore._waiters is not None else 0
                    active = self.limit_model_concurrency - available
                    queue = waiters
                    
                    model_status[model_name] = {
                        "active": active,
                        "queue": queue,
                        "available": available,
                        "limit": self.limit_model_concurrency
                    }
                    
                    total_queue += queue
                    total_active += active
                except Exception as e:
                    self.logger.warning(f"Failed to get status for model {model_name}: {e}")
                    model_status[model_name] = {
                        "active": 0,
                        "queue": 0,
                        "available": self.limit_model_concurrency,
                        "limit": self.limit_model_concurrency
                    }
            
            # 获取worker基本状态
            status_info = self.worker.get_status_info()
            
            return {
                "timestamp": time.time(),
                "total_active": total_active,
                "total_queue": total_queue,
                "total_limit": len(self.model_semaphores) * self.limit_model_concurrency,
                "model_status": model_status,
                "worker_id": self.worker.worker_id,
                # "uptime": time.time() - (status_info.start_time if status_info.start_time else time.time()),
                "model_count": len(self.worker.models),
                "global_counter": self.global_counter
            }
        except Exception as e:
            self.logger.error(f"Failed to get monitor status: {e}")
            return {
                "error": str(e),
                "timestamp": time.time()
            }
    
    
    async def shutdown_worker(self, background_tasks: BackgroundTasks):
        """接收来自Ctrl的关闭worker的信息"""
        background_tasks.add_task(self.worker.shutdown_worker)
        self.worker._is_deleted_in_controller = True
        wid = self.worker.worker_id
        return WorkerStoppedInfo(
            id=wid, stopped=True, 
            message=f"Worker `{wid}` shutdown",
            shutdown=True,
            )

    @classmethod
    def register_worker(
            cls,
            model: HRemoteModel = None,
            worker_config: HWorkerConfig = None,
            daemon: bool = False,
            standalone: bool = False,
            **kwargs,
            ):
        """注册HModelWorker到HaiDDF"""
        from .utils import run_standlone_worker_demo
        if standalone:  # 独立程序模式，用户测试程序
            # cls().logger.info("Running `worker` in standalone mode.")
            print("Running `worker` in standalone mode.")
            return run_standlone_worker_demo()
        
        assert model is not None, f"Model should be not None"
        import uvicorn
        app: FastAPI = HWorkerAPP(model, worker_config=worker_config, **kwargs)
        def run_uvicron():
            uvicorn.run(app, host=app.host, port=app.port)

        if daemon:  # 守护进程模式，app workre在后台运行
            t = threading.Thread(target=run_uvicron, daemon=True)
            t.start()
            return app.worker.get_worker_info()
        else:  # 正常模式，app worker在前台运行
            run_uvicron()

    def run(self):
        import uvicorn
        uvicorn.run(self, host=self.host, port=self.port)
    
    def __del__(self):
        """清理资源"""
        if hasattr(self, '_thread_pool') and self._thread_pool:
            self._thread_pool.shutdown(wait=False)










# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["StaticFileChunkingStrategyParam"]


class StaticFileChunkingStrategyParam(TypedDict, total=False):
    chunk_overlap_tokens: Required[int]
    """The number of tokens that overlap between chunks. The default value is `400`.

    Note that the overlap must not exceed half of `max_chunk_size_tokens`.
    """

    max_chunk_size_tokens: Required[int]
    """The maximum number of tokens in each chunk.

    The default value is `800`. The minimum value is `100` and the maximum value is
    `4096`.
    """



'''
Adapted from:
https://github.com/keras-team/keras/blob/master/keras/utils/data_utils.py
'''

import hashlib
import os
import shutil
import zipfile
import tarfile
import urllib
# import requests


def _download(url, fname, chunk_size=1024):
    from tqdm import tqdm
    import requests
    '''https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51'''
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    with open(fname, 'wb') as file, tqdm(
        desc=fname,
        total=total,
        unit='iB',
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)


def extract_archive(file_path, path='.', archive_format='auto'):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.
    Args:
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.
    Returns:
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, str):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type == 'tar':
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type == 'zip':
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def _hash_file(fpath, algorithm='md5', chunk_size=131071):
    """Calculates a file sha256 or md5 hash.
    # Example
    ```python
        >>> from keras.data_utils import _hash_file
        >>> _hash_file('/path/to/file.zip')
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```
    # Arguments
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.
    # Returns
        The file hash
    """
    if (algorithm == 'sha256') or (algorithm == 'auto'):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, 'rb') as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm='md5', chunk_size=131071):
    """Validates a file against a sha256 or md5 hash.
    # Arguments
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.
    # Returns
        Whether the file is valid
    """
    if ((algorithm == 'sha256') or (algorithm == 'auto' and len(file_hash) == 64)):
        hasher = 'sha256'
    else:
        hasher = 'md5'

    return str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash)


def get_file(origin=None,
             fname=None,
             file_hash=None,
             datadir='datasets',
             hash_algorithm='md5',
             extract=False,
             force_download=False,
             archive_format='auto'):
    """Downloads a file from a URL if it not already in the cache.
    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.
    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.
    Args:
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location. If `None`, the
            name of the file at `origin` will be used.
        origin: Original URL of the file.
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are `'md5'`, `'sha256'`, and `'auto'`.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are `'auto'`, `'tar'`, `'zip'`, and `None`.
            `'tar'` includes tar, tar.gz, and tar.bz files.
            The default `'auto'` corresponds to `['tar', 'zip']`.
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the default directory `datasets/`.
    Returns:
        Path to the downloaded file
    """
    if origin is None:
        raise ValueError('Please specify the "origin" argument (URL of the file '
                         'to download).')

    os.makedirs(datadir, exist_ok=True)

    if not fname:
        fname = os.path.basename(urllib.parse.urlsplit(origin).path)
        if not fname:
            raise ValueError(
                f"Can't parse the file name from the origin provided: '{origin}'."
                "Please specify the `fname` as the input param.")

    fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath) and not force_download:
        # File found; verify integrity if a hash was provided.
        print(f'A local file already found at {fpath}, checking hash...')
        if file_hash is not None:
            if validate_file(fpath, file_hash, algorithm=hash_algorithm):
                print('Local file hash matches, no need to download.')
            else:
                print(
                    'A local file was found, but it seems to be '
                    f'incomplete or outdated because the {hash_algorithm} '
                    f'file hash does not match the original value of {file_hash} '
                    'so we will re-download the data.')
                download = True
    else:
        download = True

    if download:
        print(f'Downloading data from {origin} to {fpath}')

        error_msg = 'URL fetch failure on {}: {}'
        import requests
        try:
            try:
                _download(origin, fpath)
            except requests.exceptions.RequestException as e:
                raise Exception(error_msg.format(origin, e.msg))
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(fpath):
                os.remove(fpath)
            raise

        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                if os.path.exists(fpath):
                    os.remove(fpath)
                raise RuntimeError(f'Checksum does not match for file {fpath}')

    if extract:
        extract_archive(fpath, datadir, archive_format)

    return fpath, download



import os, sys, signal
from dataclasses import dataclass, field, asdict
from typing import Dict, List, Optional, Union, Generator, Callable, Literal, AsyncGenerator
import threading
import atexit
import warnings
import time, asyncio
import traceback, inspect
from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.responses import StreamingResponse
import requests
import logging
from logging import Logger


from ._related_class import (
    WorkerNetworkInfo, ModelResourceInfo, WorkerStatusInfo, WorkerInfo,
    HRemoteModel
)

# from .utils import get_uuid
from . import utils

@dataclass
class HWorkerConfig:  # (2) worker的参数配置和启动代码
    # config for worker server
    host: str = field(default="0.0.0.0", metadata={"help": "Worker's address, enable to access from outside if set to `0.0.0.0`, otherwise only localhost can access"})
    port: Union[int, str, None] = field(default=42600, metadata={"help": "Worker's port, default is None, which means auto start from `auto_start_port`"})
    auto_start_port: int = field(default=42602, metadata={"help": "Worker's start port, only used when port is set to `auto`"})
    route_prefix: str = field(default="/apiv2", metadata={"help": "Route prefix for worker"})
    
    # config for controller connection
    no_register: bool = field(default=True, metadata={"help": "Do not register to controller"})
    controller_address: str = field(default="http://localhost:42601", metadata={"help": "Controller's address"})
    controller_prefix: str = field(default="/apiv2", metadata={"help": "Controller's route prefix"})
    controller_key: str = field(default="", metadata={"help": "API key for reigster to controller, ensure the security, you must provide it if the `controller` requires authentication"})
    
    # config for model
    speed: int = field(default=1, metadata={"help": "Model's speed"})
    limit_model_concurrency: int = field(default=100, metadata={"help": "Limit the model's concurrency"})
    stream_interval: float = field(default=0., metadata={"help": "Extra interval for stream response"})
    permissions: dict = field(default_factory=lambda: {'groups': ['payg']}, metadata={"help": "Model's permissions, e.g., {'groups': ['default'], 'users': ['a', 'b'], 'owner': 'c'}"})
    description: str = field(default='This is a demo worker of HEP AI framework (HepAI)', metadata={"help": "Model's description"})
    author: str = field(default="hepai", metadata={"help": "Model's author"})
    debug: bool = field(default=False, metadata={"help": "Debug mode"})
    type: Literal["llm", "actuator", "preceptor", "memory", "common"] = field(default="common", metadata={"help": "Specify worker type, could be help in some cases"})
    daemon: bool = field(default=False, metadata={"help": "Run as daemon"})
    is_free: bool = field(default=True, metadata={"help": "Whether the model is free to use, if False, model owner should setup model pricing via controller"})
    _metadata: dict = field(default_factory=dict, metadata={"help": "Additional metadata for worker/model"})
    
    # config for common features
    enable_secret_key: bool = field(default=False, metadata={"help": "Enable secret key for worker, ensure the security, if enabled, the `api_key` must be provided when someone wants to access the worker's APIs"})
    enable_llm_router: bool = field(default=False, metadata={"help": "Enable LLM router, only for llm worker"})
    model_config_dir: Optional[str] = field(default=None, metadata={"help": "Directory to store model_config.yaml, if None, will try to use worker script directory or current working directory"})
    # enable_mcp: bool = field(default=False, metadata={"help": "Enable MCP (Model Context Protocol) for LLM worker"})


    def __post_init__(self):
        if isinstance(self.port, str):
            if self.port.lower() == 'none':
                self.port = None
            else:
                try:
                    self.port = int(self.port)
                except ValueError:
                    self.port = None
        if isinstance(self.permissions, str):
            try:
                perms = dict()
                for a in self.permissions.split(';'):
                    if ':' not in a:
                        raise ValueError(f"Invalid permissions format: '{a}'")
                    user_or_group, names = map(str.strip, a.split(':', 1))
                    if user_or_group not in ['owner', 'users', 'groups']:
                        raise ValueError(f"Invalid key in permissions: '{user_or_group}'")
                    perms[user_or_group] = (
                        [name.strip() for name in names.split(',')] if ',' in names else [names.strip()]
                    )
                # 确保owner是单个字符串
                if 'owner' in perms:
                    if isinstance(perms['owner'], list):
                        if len(perms['owner']) > 1:
                            raise ValueError(f"Only one owner is allowed, but got {perms['owner']}")
                        perms['owner'] = perms['owner'][0]
                self.permissions = perms
            except Exception as e:
                raise ValueError(f"Failed to parse permissions string: {self.permissions}. Error: {e}")
        
        
    def update_from_dict(self, d: Dict):
        """更新配置"""
        for k, v in d.items():
            if hasattr(self, k):
                setattr(self, k, v)
        unknown_keys = set(d.keys()) - set(self.__dict__.keys())
        if unknown_keys:
            warnings.warn(f"[HWorkerConfig] Encountered unknown keys when updating from dict: {unknown_keys}")
        return self
    
    def to_dict(self):
        return asdict(self)
    
    @classmethod
    def from_dict(cls, d: Dict):
        """从字典创建配置实例"""
        return cls().update_from_dict(d)


class CommonWorker:
    """
    Common Worker，用来实际处理各种函数
    v2.1支持单worker搭载多模型
    """
    def __init__(self,
                 app: FastAPI,  # FastAPI app
                 models: HRemoteModel | List[HRemoteModel],  # 模型，v2.1支持
                 worker_config: "HWorkerConfig",  # worker的配置
                 logger: Optional[Logger] = None,  # logger
                 **kwargs
                 ):
        # 配置
        self.config: HWorkerConfig = worker_config
        self.config_dict: Dict = self._init_config_dict(worker_config, **kwargs)
        self.name = self.config_dict.get("name", self.__class__.__name__)
        # 网络信息相关项目，多模型共用一个
        self.network_info: WorkerNetworkInfo = self.get_network_info()
        
        # 模型资源信息相关
        self.worker_permissions: Dict = self._parse_permissions(self.config.permissions)  # 将str转换为dict

        # 状态信息相关
        self.app = app  # "HWorkerAPP" 
        self.speed = self.config_dict.get("speed", 1)  # 这是worker的
        self.status = "idle"
        self.start_time = time.time()

        # 初始化信息
        self.models: List[HRemoteModel] = self._check_models(models)
        self._model_names = [m.name for m in self.models]
        unique_model_names = set(self._model_names)
        if len(unique_model_names) != len(self._model_names):
            raise ValueError(f"Model names should be unique, but got {self._model_names}")
        
        worker_id = self.config_dict.get("worker_id", None)
        indicators = ["worker"] + self._model_names
        self.worker_id = worker_id if worker_id else utils.gen_one_id(
            lenth=15, prefix="wk-", extra_indicators=indicators
        )
        self.stream_interval = self.config_dict.get("stream_interval", 0)
        self.controller_key = self.config_dict.get("controller_key", "")
        # self.model = model or HRemoteModel()  # deprecated, v2.1支持多模型
        
        # 建立快速查找的模型映射
        self._model_map: Dict[str, HRemoteModel] = {m.name: m for m in self.models}

        # 初始化模型状态管理器
        # 确定配置文件路径：优先使用配置的目录，否则尝试从调用栈中找到 worker 脚本目录
        if self.config.model_config_dir:
            config_dir = self.config.model_config_dir
        else:
            # 尝试从调用栈中找到调用 worker 的脚本目录
            import inspect
            config_dir = None
            for frame_info in inspect.stack():
                frame_file = frame_info.filename
                # 查找名为 *_worker.py 的文件
                if frame_file.endswith('_worker.py') and 'hepai/components/haiddf' not in frame_file:
                    config_dir = os.path.dirname(os.path.abspath(frame_file))
                    break
            # 如果找不到，使用当前工作目录
            if config_dir is None:
                config_dir = os.getcwd()

        config_file = os.path.join(config_dir, "model_status.yaml")
        from .model_status_manager import ModelStatusManager
        self.model_status_manager = ModelStatusManager(worker=self, config_file=config_file)

        # flag
        self._is_deleted_in_controller = False  # a flag to indicate whether the worker is deleted in controller
        # 日志
        self.logger = logger or logging.getLogger(__name__)
        # 注册模型
        if not self.config.no_register:
            success: bool = self.register_to_controller()
            if success:  # sent hartbeat every 60s
                # 使用异步心跳以提高性能
                try:
                    loop = asyncio.get_event_loop()
                    if loop.is_running():
                        # 在异步环境下，创建后台任务
                        loop.create_task(self.worker_heartbeat_async())
                    else:
                        # 没有事件循环，回退到线程模式
                        self.heartbeat_thread = threading.Thread(
                            target=self.worker_heartbeat, 
                            daemon=True,
                            )
                        self.heartbeat_thread.start()
                except RuntimeError:
                    # 没有事件循环，回退到线程模式
                    self.heartbeat_thread = threading.Thread(
                        target=self.worker_heartbeat, 
                        daemon=True,
                        )
                    self.heartbeat_thread.start()
            else:  # if not success, flags as local worker only
                self._is_deleted_in_controller = True

        # 标识是否已经在controller中删除，如果已经删除，则exit_handler不再向controller发送删除信息
        # 用于适配controllre端主动向worker发送删除worker时
        atexit.register(self.exit_handler)

        
        self.status = "ready"

    @property
    def base_url(self):
        """The base url of controller"""
        controller_addr = self.config.controller_address
        controller_prefix = self.config.controller_prefix
        if controller_prefix:
            return f'{controller_addr}{controller_prefix}'
        return controller_addr
    
    @property
    def worker_name(self):
        """deprecated in v2.1 for multi-model"""
        raise DeprecationWarning("This property is deprecated in v2.1 for multi-model, please use `worker_id` instead.")
        # return self.model.name or self.model.__class__.__name__
        
    @property
    def model_names(self) -> List[str]:
        if not self._model_names:
            self._model_names = [m.name for m in self.models]
        return self._model_names
    
    def _check_models(self, models: List[HRemoteModel] | HRemoteModel) -> List[HRemoteModel]:
        """检查模型是否符合要求"""
        # v2.1支持多模型
        if isinstance(models, HRemoteModel):
            models = [models]
        # 检查类型
        for model in models:  # 是HRemoteModel的实例或者子类
            if isinstance(model, HRemoteModel):
                continue
            elif issubclass(model, HRemoteModel):
                continue
            else:
                raise ValueError(f"The Model should be HRemoteModel or subclass of HRemoteModel, i.e. from hepai import HRemoteModel, but got {type(model)}")
        # 检查模型不重名
        model_names = [m.name for m in models]
        if len(model_names) != len(set(model_names)):
            raise ValueError(f"Model names should be unique, but got {model_names}")
        return models
    
    def _parse_permissions(self, permissions: Optional[Union[str, Dict]]) -> Dict:
        """worker授予用户或者组的权限"""
        if permissions is None:
            return {}
        elif isinstance(permissions, dict):
            return permissions
        elif isinstance(permissions, str):
            """user: <user1>; user: <user2>; group: <group1>, ..."""
            perms = dict()
            for a in permissions.split(';'):
                user_or_group, names = map(str.strip, a.split(':', 1))
                assert user_or_group in ['owner', 'users', 'groups'], f"Invalid key: {user_or_group}"
                perms[user_or_group] = (
                    [name.strip() for name in names.split(',')] if ',' in names else [names.strip()]
                )
            # 确保owner是单个字符串
            if 'owner' in perms:
                if isinstance(perms['owner'], list):
                    if len(perms['owner']) > 1:
                        raise ValueError(f"Only one owner is allowed, but got {perms['owner']}")
                    perms['owner'] = perms['owner'][0]
        else:
            raise ValueError(f"permissions should be str or dict, but got {type(permissions)}")
        return perms
    
    def _init_config_dict(self, config: HWorkerConfig, **kwargs):
        config_dict = config.__dict__
        for k, v in kwargs.items():
            if k in config_dict:
                raise ValueError(f"Duplicate key: {k}, Please check the config class and the kwargs.")
            config_dict[k] = v
        return config_dict
    
    async def worker_heartbeat_async(self):
        """异步心跳函数，避免阻塞"""
        heartbeat_interval = self.config_dict.get("heartbeat_interval", 60)
        while True:
            await asyncio.sleep(heartbeat_interval)
            # 使用线程池执行同步的注册操作
            loop = asyncio.get_event_loop()
            try:
                await loop.run_in_executor(None, self.register_to_controller, True)
            except Exception as e:
                self.logger.error(f"Heartbeat failed: {e}")
    
    def worker_heartbeat(self):
        """间隔发送心跳的函数 - 保留用于向后兼容"""
        heartbeat_interval = self.config_dict.get("heartbeat_interval", 60)
        while True:
            time.sleep(heartbeat_interval)
            self.register_to_controller(heartbeat_flag=True)
    
    def get_network_info(self) -> WorkerNetworkInfo:
        host = self.config.host
        from .utils import auto_port, auto_worker_address, get_hostname
        port = auto_port(port=self.config.port, start=self.config.auto_start_port)
        worker_address = auto_worker_address(worker_address="auto", host=host, port=port)
        route_prefix = self.config.route_prefix
        if route_prefix:
            worker_address = f"{worker_address}{route_prefix}"
        return WorkerNetworkInfo(
            host=self.config.host,
            port=port,
            host_name=get_hostname(),
            route_prefix=route_prefix,
            worker_address=worker_address,
        )
    
    def get_model_resource_info(self) -> List[ModelResourceInfo]:
        """
        v2.1 支持多模型
        v2.2 过滤禁用的模型（用于心跳上报）
        """

        model_resources = []
        for model in self.models:
            # 跳过禁用的模型
            if not self.model_status_manager.is_model_enabled(model.name):
                continue

            model_name = model.name or model.__class__.__name__
            permission = model.permission
            if not permission:  # 如果没有设置权限，则使用worker的权限
                permission = self.worker_permissions
            owner = permission.get("owner", None)
            owner = owner if owner else self.config.author

            mr = ModelResourceInfo(
                model_name=model_name,
                model_type=self.config_dict.get("model_type", "common"),
                model_version=self.config_dict.get("model_version", "1.0"),
                model_description=self.config.description,
                model_author=self.config.author,
                model_owner=permission.get("owner", None),
                model_users=permission.get("users", []),
                model_groups=permission.get("groups", []),
                model_functions=model.all_remote_callables,
                id=model.model_id,
                created=model.created,
                owned_by=owner,
            )
            model_resources.append(mr)
        return model_resources
    
    def get_callable_functions(self, cls):
        """获取一个类的所有可调用函数"""
        callable_funcs = []
        for name, method in inspect.getmembers(cls, inspect.isfunction):
            # Check if method belongs to the class and is not inherited
            if method.__qualname__.startswith(cls.__name__ + '.'):
                callable_funcs.append(name)
        return callable_funcs
    
    def get_all_callable_functions(self, cls: HRemoteModel):
        """获取一个类的所有可调用函数，包括父类的"""
        # Recursive function to collect callable methods from the class and its parent classes
        methods = {}
        # Traverse the method resolution order (MRO), which includes the class and its bases
        for subclass in inspect.getmro(cls.__class__):
            for name, method in inspect.getmembers(subclass, inspect.isfunction):
                methods[name] = method
        callable_funcs = list(methods.keys())

        # 再判断这些函数是不是可以远程调用的
        callable_funcs = [f for f in callable_funcs if hasattr(methods[f], "is_remote_callable")]

        return callable_funcs


    def get_status_info(self) -> WorkerStatusInfo:
        return WorkerStatusInfo(
            speed=self.speed,
            queue_length=self.app.get_queue_length(),
            status=self.status,
            start_time=self.start_time,
        )
    
    def get_model_by_name(self, model_name: str) -> Optional[HRemoteModel]:
        """快速获取模型对象"""
        return self._model_map.get(model_name)
    
    async def shutdown_worker(self) -> None:
        await asyncio.sleep(1)
        os.kill(os.getpid(), signal.SIGINT)
    
    def get_worker_info(self) -> WorkerInfo:
        """
        Dynamically get worker info
        """
        status_info = self.get_status_info()
        
        metadata = {
            "description": self.config.description, 
            "author": self.config.author,
            "limit_model_concurrency": self.config.limit_model_concurrency,
            "permissions": self.worker_permissions,
            "uptime": time.time() - (status_info.start_time if status_info.start_time else time.time()),
            "is_free": self.config.is_free,
        }
        worker_meta = self.config._metadata
        metadata.update(worker_meta)
        
        worker_info = WorkerInfo(
            id=self.worker_id,
            type=self.config.type,
            network_info=self.network_info,
            resource_info=self.get_model_resource_info(),
            status_info=status_info,
            check_heartbeat=True,
            version=self.config_dict.get("version", "2.0"),
            metadata=metadata,
        )
        return worker_info
    
    def set_worker_pressions(self, permissions: Dict):
        """设置模型的权限, 用于动态更新，仅限于owner和admin"""
        self.worker_permissions = permissions

    def set_worker_speed(self, speed: int):
        """设置模型的速度, 用于动态更新，仅限于owner和admin"""
        self.speed = speed
    
    ### --- 此处开始是向controller发送请求的相关函数 --- ###
    @property
    def headers(self) -> dict:
        headers = {
            'Authorization': f'Bearer {self.controller_key}' if self.controller_key else '',
            'Content-Type': 'application/json'
        }
        return headers
    
    def register_to_controller(self, heartbeat_flag: bool = False):
        """注册和心跳"""
        # logger.info(f'Register model "{self.model_name}" to controller.')
        
        # url = self.base_url + '/register_worker'
        url = self.base_url + '/worker/register_worker'
        worker_info: WorkerInfo = self.get_worker_info()
        data = worker_info.to_dict()
        
        try:
            r = requests.post(url, json=data, headers=self.headers)
            if r.status_code != 200:
                raise ValueError(f"Register worker failed. Error: {r.text}\n  Request url: {url}\n  Worker_info: {worker_info}")
        except Exception as e:
            if heartbeat_flag:
                print(f"Sent heartbeat failed, ignore... url: {url}, worker_id: {self.worker_id}")
                pass
            else:
                if self.config.debug:
                    # raise f'Register worker to controller failed. Error: {e}'
                    raise ValueError(f'Register worker to controller failed. Error: {e}')
                self.logger.warning(f"Register worker to controller failed, pass...")
                return False
        if heartbeat_flag:
            if self.config.debug:
                self.logger.info(f"Heartbeat sent successfully: `{worker_info.id}`")
        else:
            self.logger.info(f"Worker `{worker_info.id}` register to `{self.base_url}` successfully.")
        return True
  
    def exit_handler(self):
        """
        退出时向Contoller发送退出信息
        """
        if self.config.no_register:  # 初始化时没有注册到controller，不需要发送退出信息
            return
        if self._is_deleted_in_controller:
            return
        # logger.info(f'Remove model "{self.model_name}" from controller {self.controller_addr}, ')
        # url = self.base_url + "/stop_worker"
        url = self.base_url + "/worker/stop_worker"
        
        worker_info: WorkerInfo = self.get_worker_info()
        wid = worker_info.id
        payload = {"worker_id": wid, "permanent": False}
        # data = worker_info.to_dict()
        r = requests.post(
            url, 
            json=payload, 
            headers=self.headers)
        assert r.status_code == 200, f"Stop worker failed. {r.text}\n worker_info: {worker_info}"
        # logger.info(f'Done. {r.text}') 
        print(f"Stop worker successful. res: {r.text}")

    ### --- 此处是处理Controller的请求的相关函数 --- ###
    def unified_gate(
            self, 
            model: str,
            function: str,
            args: List = None,
            kwargs: Dict = None,
            ):
        """
        统一入口, 
        v2.0更新为符合FastAPI规范，易于检测错误
        v2.1支持多模型
        """
        # assert "function" in kwargs, "function is required"
        # function = kwargs.pop("function")

        if model is None:
            if len(self.models) == 1:
                # 不指定模型，且worker只搭载一个模型时，直接使用这个模型
                req_model: HRemoteModel = self.models[0]  # requested model
            else:
                raise HTTPException(status_code=400, detail=f"Model is required, but got None, and there are {len(self.models)} models in the worker `{self.worker_id}`")  
        else:
            # 使用快速查找替代列表推导式
            req_model: HRemoteModel = self._model_map.get(model)
            if req_model is None:
                raise HTTPException(status_code=404, detail=f"Model `{model}` does not exist in the worker `{self.worker_id}`")
                
        has_function = hasattr(req_model, function)
        if not has_function:
            raise HTTPException(status_code=404, detail=f"Function `{function}` does not exist in the worker `{self.worker_id}`")
        is_callable = callable(getattr(req_model, function))
        is_remote_callable = hasattr(getattr(req_model, function), "is_remote_callable")
        if not is_callable:
            raise HTTPException(status_code=405, detail=f"Function `{function}` is not callable in the worker `{self.worker_id}`")
        if not is_remote_callable:
            raise HTTPException(status_code=405, detail=f"Function `{function}` is not remote callable in the worker `{self.worker_id}`, please add `@remote_callable` decorator to the function in worker model.")
        if has_function and is_callable and is_remote_callable:
            func: Callable = getattr(req_model, function)
            try:
                # res =  func(**kwargs)
                try:
                    res = func(*args, **kwargs)
                except:
                    # 有时候因为中间层额外引入了stream参数，而一些函数不允许接收stream参数。
                    stream = kwargs.pop("stream", False)  # 为了在客户端传输stream时，不会被kwargs接收，所以pop出来
                    res = func(*args, **kwargs)
                    
                if isinstance(res, Generator):
                    return StreamingResponse(res, media_type="application/octet-stream")
                    # return StreamingResponse(res, media_type="text/event-stream")
                return res
            except Exception as e:
                # 获取报错类型：e.__class__.__name__
                tb_str = traceback.format_exception(*sys.exc_info())
                tb_str = "".join(tb_str)
                print(f"Error: {e}.\nTraceback: {tb_str}")
                e_class = e.__class__.__name__
                error_msg = e.__dict__.get("body", None)
                error_msg = error_msg if error_msg else f"{e_class}: {str(e)}"
                if e_class == "ModuleNotFoundError":
                    raise HTTPException(status_code=404, detail=error_msg)
                elif e_class == "NotImplementedError":
                    raise HTTPException(status_code=501, detail=error_msg)
                elif e_class == "BadRequestError":
                    raise HTTPException(status_code=400, detail=error_msg)
                elif e_class == "TypeError":
                    raise HTTPException(status_code=500, detail=error_msg)
                elif e_class == "APITimeoutError":
                    raise HTTPException(status_code=504, detail=error_msg)
                ## TODO: 其他报错类型转换为合适的报错状态码
                error_msg2 = f"{e_class}: {str(e)}"
                print(f"一种新的错误类型：{e_class}, 错误信息：{error_msg}\n{error_msg2}")
                raise HTTPException(status_code=400, detail=f'{error_msg2}\n{error_msg2}')
        else:
            raise HTTPException(status_code=404, detail=f"Function `{function}` does not exist or is not callable in the worker `{self.worker_id}`")
        
    async def unified_gate_async(
            self, 
            model: str,
            function: str,
            args: List = None,
            kwargs: Dict = None,
            ):
        """
        统一入口, 
        v2.0更新为符合FastAPI规范，易于检测错误
        v2.1支持多模型
        """
        # assert "function" in kwargs, "function is required"
        # function = kwargs.pop("function")

        if model is None:
            if len(self.models) == 1:
                # 不指定模型，且worker只搭载一个模型时，直接使用这个模型
                req_model: HRemoteModel = self.models[0]  # requested model
            else:
                raise HTTPException(status_code=400, detail=f"Model is required, but got None, and there are {len(self.models)} models in the worker `{self.worker_id}`")  
        else:
            # 使用快速查找替代列表推导式
            req_model: HRemoteModel = self._model_map.get(model)
            if req_model is None:
                raise HTTPException(status_code=404, detail=f"Model `{model}` does not exist in the worker `{self.worker_id}`")
                
        has_function = hasattr(req_model, function)
        if not has_function:
            raise HTTPException(status_code=404, detail=f"Function `{function}` does not exist in the worker `{self.worker_id}`")
        is_callable = callable(getattr(req_model, function))
        is_remote_callable = hasattr(getattr(req_model, function), "is_remote_callable")
        if not is_callable:
            raise HTTPException(status_code=405, detail=f"Function `{function}` is not callable in the worker `{self.worker_id}`")
        if not is_remote_callable:
            raise HTTPException(status_code=405, detail=f"Function `{function}` is not remote callable in the worker `{self.worker_id}`, please add `@remote_callable` decorator to the function in worker model.")
        if has_function and is_callable and is_remote_callable:
            func: Callable = getattr(req_model, function)
            # 判断是否为异步函数
            is_async = asyncio.iscoroutinefunction(func)
            try:
                if is_async:
                    res = await func(*args, **kwargs)
                    # try:
                    #     res = await func(*args, **kwargs)
                    # except Exception as e:
                    #     raise RuntimeError(f"Async function raised an error, please check the function. {e}")
                    # 判断是否是异步流式响应
                    if isinstance(res, AsyncGenerator):
                        #返回异步流式响应
                        return StreamingResponse(res, media_type="application/octet-stream")
                        # return StreamingResponse(res, media_type="text/event-stream") 
                    return res  
                else:
                    # res =  func(**kwargs)
                    try:
                        res = func(*args, **kwargs)
                    except:
                        # 有时候因为中间层额外引入了stream参数，而一些函数不允许接收stream参数。
                        stream = kwargs.pop("stream", False)  # 为了在客户端传输stream时，不会被kwargs接收，所以pop出来
                        res = func(*args, **kwargs)
                        
                    if isinstance(res, Generator):
                        return StreamingResponse(res, media_type="application/octet-stream")
                        # return StreamingResponse(res, media_type="text/event-stream")
                    return res
            except Exception as e:
                # 获取报错类型：e.__class__.__name__
                tb_str = traceback.format_exception(*sys.exc_info())
                tb_str = "".join(tb_str)
                print(f"[CommonWorker]Error: {e}.\nTraceback: {tb_str}")
                e_class = e.__class__.__name__
                error_msg = e.__dict__.get("body", None)
                error_msg = error_msg if error_msg else f"{e_class}: {str(e)}"
                if e_class == "ModuleNotFoundError":
                    raise HTTPException(status_code=404, detail=error_msg)
                elif e_class == "NotImplementedError":
                    raise HTTPException(status_code=501, detail=error_msg)
                elif e_class == "BadRequestError":
                    raise HTTPException(status_code=400, detail=error_msg)
                elif e_class == "TypeError":
                    raise HTTPException(status_code=500, detail=error_msg)
                elif e_class == "APITimeoutError":
                    raise HTTPException(status_code=504, detail=error_msg)
                ## TODO: 其他报错类型转换为合适的报错状态码
                error_msg2 = f"{e_class}: {str(e)}"
                print(f"[CommonWorker]一种新的错误类型：{e_class}, 错误信息：{error_msg}\n{error_msg2}")
                raise HTTPException(status_code=400, detail=f'{error_msg2}\n{error_msg2}')
                # raise e
        else:
            raise HTTPException(status_code=404, detail=f"Function `{function}` does not exist or is not callable in the worker `{self.worker_id}`")




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Annotated, TypeAlias

from ...._utils import PropertyInfo
from .text_delta_block import TextDeltaBlock
from .refusal_delta_block import RefusalDeltaBlock
from .image_url_delta_block import ImageURLDeltaBlock
from .image_file_delta_block import ImageFileDeltaBlock

__all__ = ["MessageContentDelta"]

MessageContentDelta: TypeAlias = Annotated[
    Union[ImageFileDeltaBlock, TextDeltaBlock, RefusalDeltaBlock, ImageURLDeltaBlock],
    PropertyInfo(discriminator="type"),
]





"""
    ### anthropic: v0.59.0
支持claude code 2.0.5，但不支持2.0.62
报错：API Error: 400 {"detail":"RuntimeError: Async function raised an error, please check the function. AsyncMessages.create() got an unexpected keyword argument 'context_management'\nRuntimeError: Async function raised an error, please check the function. AsyncMessages.create() got an unexpected 
    keyword argument 'context_management'"}
原因是：0.59.0版本的anthropic包不支持context_management参数

"""
from typing import List, Optional


from hepai.workers.zhizz.zhizz_worker import ZhizzModelConfig, ZhizzWorkerConfig
from hepai.components.haiddf.base_class._llm_remote_model import LLMRemoteModel, LLMModelConfig
from hepai.workers.zhizz.utils import load_models

import uvicorn
from fastapi import FastAPI
import hepai as hai
    
model_config, worker_config = hai.parse_args((ZhizzModelConfig, ZhizzWorkerConfig))
models: List[LLMRemoteModel] = load_models(model_config)  # Load models from the configuration file.

model = [x for x in models if x.config.name == "claude-sonnet-4-5-20250929"][0]


async def main():
    
    
    
    gen = await model.anthropic_stream(
        model="claude-sonnet-4-5-20250929",
        messages=[{"role": "user", "content": "写一首关于春天的诗"}],
        max_tokens=1024,
        stream=True,
    )







"""
python的argparse模块不能使用两个，
例如：在大型项目中，某些子模块使用了argparse来解析参数，而主程序也使用了argparse，由于主程序引用了子程序，子程序先被加载，
此时只有子程序的argparse可用，而主程序的argparse不可用，这样就会导致主程序的argparse解析出错。
解决方案：damei库提供一个假的argparse，拥有与系统的argparse相同的函数，但是不会解析命令行参数
在子程序中:
    parser = dm.argparse.ArgumentParser() 来替代parser = argparse.ArgumentParser()

"""
from .config_loader import PyConfigLoader as Config


class Args(object):
    def __init__(self, *args, **kwargs):
        """
        properties:
            name: e.g. source
            short_name: e.g. s
            help: e.g. source help
            # default: e.g. default value
            value: e.g. value
            type: e.g. str
        """

        for arg in args:
            if arg.startswith('--'):
                self.name = arg[2:]
            elif arg.startswith('-'):
                self.short_name = arg[1:]

        # self.short_name = kwargs.get('short_name', None)
        self.nargs = kwargs.get('nargs', None)
        self._type = kwargs.get('type', None)
        self.default = kwargs.get('default', None)
        self.help = kwargs.get('help', None)
        self.action = kwargs.get('action', None)

    @property
    def type(self):
        if self.nargs is None:
            return self._type
        else:
            return list

    @property
    def value(self):
        if self.action == 'store_true':
            return False
        elif self.action == 'store_false':
            return True
        elif self.type == bool:
            return bool(self.default)
        elif self.type == int:
            if self.default is None:
                return None
            return int(self.default)
        elif self.type == float:
            return float(self.default)
        elif self.type == str:
            ret = self.default if self.default is None else str(self.default)
            return ret
        else:
            return self.default


class ArgumentParser(object):
    """TODO: 还有一些函数没完善，例如add_argument_group"""

    def __init__(self, *args, **kwargs):
        self._args = []
        self._args_dict = {}
        self.prog = kwargs.get('prog', None)
        self.description = kwargs.get('description', None)

    def __repr__(self):
        stored_args = []
        for k, v in self._args_dict.items():
            k = k.replace('-', '_')
            v2 = getattr(self, k)
            stored_args.append(f'{k}={v2}')  # 如果直接返回v，则使用opt.attr = new_value复制后，不显示新的值，改成v2，则显示新的参数
        stored_args = ', '.join(stored_args)
        format_str = f'Namespace({stored_args}) in {self.__class__}'
        return format_str

    def __dict__(self):
        return self._args_dict

    @property
    def args(self):
        return self._args

    def parse_args(self):
        name = self.prog if self.prog else 'cfg_from_argparser'
        return Config.from_argparser(self, name=name)

    def parse_known_args(self):
        return self.args, []

    def add_argument(self, *args, **kwargs):
        # print('x2x')
        arg = Args(*args, **kwargs)
        self._args.append(arg)  # object
        name = arg.name
        if '-' in name:
            name = name.replace('-', '_')
        self._args_dict[name] = arg.value

        setattr(self, name, arg.value)      
        # print(name)

    def add_argument_group(self, *args, **kwargs):
        pass

    def add_mutually_exclusive_group(self, *args, **kwargs):
        pass

    def add_subparsers(self, *args, **kwargs):
        pass

    def add_help(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass

    def exit(self, *args, **kwargs):
        pass

    def print_help(self, *args, **kwargs):
        pass

    def print_usage(self, *args, **kwargs):
        pass

    def print_version(self, *args, **kwargs):
        pass

    def set_defaults(self, *args, **kwargs):
        pass

    def set_description(self, *args, **kwargs):
        pass

    def set_epilog(self, *args, **kwargs):
        pass

    def set_usage(self, *args, **kwargs):
        pass


# def ArgumentParser(*args, **kwargs):
    # return FakeArgparse(*args, **kwargs)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from ...._compat import cached_property
from .completions import Completions, AsyncCompletions
from ...._resource import SyncAPIResource, AsyncAPIResource

__all__ = ["Chat", "AsyncChat"]


class Chat(SyncAPIResource):
    @cached_property
    def completions(self) -> Completions:
        return Completions(self._client)


class AsyncChat(AsyncAPIResource):
    @cached_property
    def completions(self) -> AsyncCompletions:
        return AsyncCompletions(self._client)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from typing_extensions import Literal

from pydantic import Field as FieldInfo

from ..._models import BaseModel

__all__ = ["ResponseFormatJSONSchema", "JSONSchema"]


class JSONSchema(BaseModel):
    name: str
    """The name of the response format.

    Must be a-z, A-Z, 0-9, or contain underscores and dashes, with a maximum length
    of 64.
    """

    description: Optional[str] = None
    """
    A description of what the response format is for, used by the model to determine
    how to respond in the format.
    """

    schema_: Optional[Dict[str, object]] = FieldInfo(alias="schema", default=None)
    """The schema for the response format, described as a JSON Schema object."""

    strict: Optional[bool] = None
    """Whether to enable strict schema adherence when generating the output.

    If set to true, the model will always follow the exact schema defined in the
    `schema` field. Only a subset of JSON Schema is supported when `strict` is
    `true`. To learn more, read the
    [Structured Outputs guide](https://platform.openai.com/docs/guides/structured-outputs).
    """


class ResponseFormatJSONSchema(BaseModel):
    json_schema: JSONSchema

    type: Literal["json_schema"]
    """The type of response format being defined: `json_schema`"""



from typing import List, Callable, Dict
from dataclasses import field, dataclass
from fastapi import APIRouter, Depends, HTTPException
from fastapi.requests import Request
from fastapi import FastAPI

from .._worker_class import CommonWorker
from ..utils import(
    read_request_body
    )
from ..worker_app import HWorkerAPP, FunctionParamsItem
from ..singletons import authorizer

from hepai.workers.zhizz.utils import get_provider_by_model_name

api_key_auth: Callable = Depends(authorizer.api_key_auth)


@dataclass
class AnthropicRouterGroup:
    name: str = "anthropic"
    prefix: str = "/apiv2"
    tags: List[str] = field(default_factory=lambda: ["anthropic"])
    router: APIRouter = field(default_factory=APIRouter)
    parent_app: HWorkerAPP = None  # type: ignore

    def __post_init__(self):
        
        self.count = 0
        rt = self.router
        rt.post("/anthropic/messages", dependencies=[api_key_auth])(self.anthropic_messages)
        rt.post("/anthropic/v1/messages", dependencies=[api_key_auth])(self.anthropic_messages)
        rt.post("/anthropic/v1/messages/count_tokens", dependencies=[api_key_auth])(self.count_tokens)

        rt.post("/anthropic/api/event_logging/batch", dependencies=[api_key_auth])(self.event_logging_batch)
        rt.post("/anthropic//api/event_logging/batch", dependencies=[api_key_auth])(self.event_logging_batch)

    async def event_logging_batch(self, request: Request):
        request_body: Dict = await read_request_body(request=request)
        # print(f"Received event logging batch:")
        # import json
        # print(json.dumps(request_body, indent=2), flush=True)
        pass

    async def anthropic_messages(self, request: Request):
        request_body: Dict = await read_request_body(request=request)
        if "model" not in request_body:
            raise HTTPException(status_code=400, detail="[AnthropicRouterGroup] This `model` must be specified")
        
        # Auto rename
        model = request_body["model"]
        if '/' not in model:
            model_name = f'{get_provider_by_model_name(model)}/{model}'
        else:
            model_name = model
        
        self.count += 1
        func_params = FunctionParamsItem(
            args=[],
            kwargs=request_body
        )
        rst = await self.parent_app.worker_unified_gate(
            function_params=func_params,
            model=model_name, 
            function="anthropic_messages",
        )
        return rst

    async def count_tokens(self, request: Request):
        request_body: Dict = await read_request_body(request=request)
        if "model" not in request_body:
            raise HTTPException(status_code=400, detail="[AnthropicRouterGroup] This `model` must be specified")
        model = request_body["model"]
        
        # Auto rename
        model = request_body["model"]
        if '/' not in model:
            model_name = f'{get_provider_by_model_name(model)}/{model}'
        else:
            model_name = model
        

        self.count += 1
        func_params = FunctionParamsItem(
            args=[],
            kwargs=request_body
        )
        rst = await self.parent_app.worker_unified_gate(
            function_params=func_params,
            model=model_name,
            function="anthropic_count_tokens",
        )
        return rst


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.


from .fine_tuning_job_wandb_integration_object import FineTuningJobWandbIntegrationObject

FineTuningJobIntegration = FineTuningJobWandbIntegrationObject



from typing import Dict, List
import time
import json


async def convert_claude_to_openai_format(response: Dict) -> Dict:
    """将claude系列模型的返回结果转换为OpenAI格式"""
    # 这里根据实际的claude模型返回结果结构进行适配
    
    # 转换msg
    choices = []
    for i, msg in enumerate(response.get("content")):
        oai_msg = {
            "index": i,
            "message": {
                "role": response.get("role", "user"),
                "content": msg.get("text", ""),
                "refusal": None,
                "annotations": msg.get("citations"),
            },
            "logprobs": None,
            "finish_reason": response.get("finish_reason"),
        }
        choices.append(oai_msg)
    
    # 不转换usage
  
    openai_format = {
        "id": response.get("id"),
        "object": "chat.completion",
        "created": time.time(),
        "model": response.get("model"),
        "choices": choices,
        "usage": response.get("usage", {}),
        "service_tier": response.get("usage", {}).get("service_tier", None)
    }
    return openai_format


async def convert_object_openai_to_anthropic(openai_obj):
    from openai.types.chat import (
        ChatCompletion, ChatCompletionMessageToolCall
    )
    from openai.types import CompletionUsage
 
    assert isinstance(openai_obj, ChatCompletion), "openai_obj must be an instance of ChatCompletion"
    
    from anthropic.types import (
        Usage, Message, 
        TextBlock, ThinkingBlock, RedactedThinkingBlock, ToolUseBlock,
        ServerToolUseBlock, WebSearchToolResultBlock
    )

    role = None
    content = []
    stop_reason = None

    
    choices = openai_obj.choices
    for choice in choices:
        finish_reason = choice.finish_reason
        msg = choice.message
        if role is None:
            role = msg.role

        if finish_reason == "tool_calls":
            stop_reason = "tool_use"
            tool_calls: List[ChatCompletionMessageToolCall] = msg.tool_calls
            for tool_call in tool_calls:
                tool_use_block = ToolUseBlock(
                    id=tool_call.id,
                    input=tool_call.function.arguments,
                    name=tool_call.function.name,
                    type="tool_use"  # function -> tool_use
                )
                content.append(tool_use_block)
        elif finish_reason == "stop":
            stop_reason = "end_turn"
            text_block = TextBlock(
                text=msg.content,
                citations=msg.annotations,
                type="text"
            )
            content.append(text_block)
        else:
            raise NotImplementedError("Only tool_calls finish_reason is implemented in this conversion.")

    
    usage: CompletionUsage = openai_obj.usage
    
    anthropic_usage = Usage(
        cache_creation_input_tokens=usage.prompt_tokens_details.cached_tokens if usage and usage.prompt_tokens_details else None,
        cache_read_input_tokens=None,
        input_tokens=usage.prompt_tokens if usage else None,
        output_tokens=usage.completion_tokens if usage else None,
        server_tool_use=None,
        service_tier=None
    )
    
    message = Message(
        id=openai_obj.id,
        content=content,
        model=openai_obj.model,
        role=role,
        stop_reason=stop_reason,
        stop_sequence=None,
        type="message",
        usage=anthropic_usage
    )
    
    return message


async def convert_openai_to_anthropic_format(response: Dict) -> Dict:
    """
    将OpenAI格式的返回结果转换为Anthropic(Claude)格式
    """
    # 解析choices
    content = []
    role = None
    finish_reason = None

    choices = response.get("choices", [])
    for choice in choices:
        msg = choice.get("message", {})
        # 只取第一个role
        if role is None:
            role = msg.get("role", "user")
        finish_reason = choice.get("finish_reason")
        content.append({
            "text": msg.get("content", ""),
            "citations": msg.get("annotations", None),
            "type": "text",
        })
        
    # 转换usage
    usage = response.get("usage", {})
    # "# cache_creation_input_tokens=None, cache_read_input_tokens=None, input_tokens=None, output_tokens=None, server_tool_use=None, service_tier=None, c"
    anthropic_usage = {
        "cache_creation_input_tokens": None,
        "cache_read_input_tokens": None,
        "input_tokens": usage.get("prompt_tokens", None),
        "output_tokens": usage.get("completion_tokens", None),
        "server_tool_use": None,
        "service_tier": response.get("service_tier", None),
    }

    anthropic_format = {
        "id": response.get("id"),
        "model": response.get("model"),
        "role": role,
        "content": content,
        "stop_reason": finish_reason,
        "stop_sequence": None,
        "type": "message",
        "usage": anthropic_usage
    }
    return anthropic_format


async def convert_openai_to_anthropic_format_stream(generator, return_dict=True):

    if return_dict:
        async for chunk in convert_openai_to_anthropic_format_stream_chunk(generator):
            yield {
                "event": chunk.get("type"),
                "data": chunk
            }
    else:
        async for chunk in convert_openai_to_anthropic_format_stream_chunk(generator):
            yield f"event: {chunk.get('type')}\ndata: {json.dumps(chunk)}\n\n"

async def convert_openai_to_anthropic_format_stream_chunk(generator):
    """将 OpenAI 流式输出转换为 Anthropic 格式"""

    first_chunk = True
    accumulated_content = ""
    message_id = None
    model = None
    usage_info = None
    content_finished = False
    tool_calls_data = {}  # 存储工具调用数据 {index: {id, name, arguments}}
    current_content_index = 0

    # i = 0
    async for chunk in generator:
        if not chunk:
            continue
        # print(f"{i}: {chunk}")
        # i += 1
        if not isinstance(chunk, dict):
            chunk = chunk.model_dump()

        # 提取基本信息
        if message_id is None:
            message_id = chunk.get('id')
        if model is None:
            model = chunk.get('model')

        # 保存 usage 信息
        if chunk.get('usage'):
            usage_info = chunk.get('usage')

        choices = chunk.get('choices', [])
        if not choices:
            # 如果没有 choices，但有 usage，说明是最后的 usage chunk
            if usage_info and (content_finished or tool_calls_data):
                # 确定停止原因
                stop_reason = 'tool_use' if tool_calls_data else 'end_turn'

                yield {
                    'delta': {
                        'stop_reason': stop_reason,
                        'stop_sequence': None
                    },
                    'type': 'message_delta',
                    'usage': {
                        'cache_creation_input_tokens': 0,
                        'cache_read_input_tokens': 0,
                        'input_tokens': usage_info.get('prompt_tokens', 0),
                        'output_tokens': usage_info.get('completion_tokens', 0),
                        'server_tool_use': None
                    }
                }

                # 构建内容列表
                content_list = []
                if accumulated_content:
                    content_list.append({
                        'citations': None,
                        'text': accumulated_content,
                        'type': 'text'
                    })

                # 添加工具调用内容
                for tool_data in tool_calls_data.values():
                    try:
                        # 尝试解析 JSON 参数
                        parsed_args = json.loads(tool_data['arguments']) if tool_data['arguments'] else {}
                    except:
                        parsed_args = tool_data['arguments']

                    content_list.append({
                        'id': tool_data['id'],
                        'input': parsed_args,
                        'name': tool_data['name'],
                        'type': 'tool_use'
                    })

                yield {
                    'type': 'message_stop',
                    'message': {
                        'id': message_id,
                        'content': content_list,
                        'model': model,
                        'role': 'assistant',
                        'stop_reason': stop_reason,
                        'stop_sequence': None,
                        'type': 'message',
                        'usage': {
                            'cache_creation_input_tokens': 0,
                            'cache_read_input_tokens': 0,
                            'input_tokens': usage_info.get('prompt_tokens', 0),
                            'output_tokens': usage_info.get('completion_tokens', 0),
                            'server_tool_use': None,
                            'service_tier': None
                        }
                    }
                }
            continue

        choice = choices[0]  # 只处理第一个choice
        delta = choice.get('delta', {})
        delta = {} if delta == [] else delta  # 有的时候delta是空列表
        finish_reason = choice.get('finish_reason')

        # 第一个chunk：发送 message_start
        if first_chunk:
            yield {
                'message': {
                    'id': message_id,
                    'content': [],
                    'model': model,
                    'role': 'assistant',
                    'stop_reason': None,
                    'stop_sequence': None,
                    'type': 'message',
                    'usage': {
                        'cache_creation_input_tokens': 0,
                        'cache_read_input_tokens': 0,
                        'input_tokens': 0,
                        'output_tokens': 0,
                        'server_tool_use': None,
                        'service_tier': None
                    }
                },
                'type': 'message_start'
            }
            first_chunk = False

        # 处理内容增量
        content = delta.get('content', '')
        if content and not tool_calls_data:  # 如果有工具调用就不处理文本内容
            if accumulated_content == "":  # 第一次文本内容
                yield {
                    'content_block': {
                        'citations': None,
                        'text': '',
                        'type': 'text'
                    },
                    'index': current_content_index,
                    'type': 'content_block_start'
                }

            accumulated_content += content

            yield {
                'delta': {
                    'text': content,
                    'type': 'text_delta'
                },
                'index': current_content_index,
                'type': 'content_block_delta'
            }

            yield {
                'type': 'text',
                'text': content,
                'snapshot': accumulated_content
            }

        # 处理tool_calls增量
        tool_calls = delta.get('tool_calls', [])
        if tool_calls is not None:
            for tool_call in tool_calls:
                tool_index = tool_call.get('index', 0)
                tool_id = tool_call.get('id')
                function_info = tool_call.get('function', {})

                # 初始化工具调用数据
                if tool_index not in tool_calls_data:
                    tool_calls_data[tool_index] = {
                        'id': tool_id or '',
                        'name': function_info.get('name', ''),
                        'arguments': ''
                    }

                    # 发送 content_block_start 事件
                    content_index = len(tool_calls_data) - 1 + (1 if accumulated_content else 0)
                    yield {
                        'content_block': {
                            'id': tool_calls_data[tool_index]['id'],
                            'input': {},
                            'name': tool_calls_data[tool_index]['name'],
                            'type': 'tool_use'
                        },
                        'index': content_index,
                        'type': 'content_block_start'
                    }

                # 更新工具调用数据
                if tool_id:
                    tool_calls_data[tool_index]['id'] = tool_id
                if function_info.get('name'):
                    tool_calls_data[tool_index]['name'] = function_info.get('name')
                if function_info.get('arguments'):
                    tool_calls_data[tool_index]['arguments'] += function_info.get('arguments')

                    # 发送 input_json 增量事件
                    content_index = len(tool_calls_data) - 1 + (1 if accumulated_content else 0)

                    yield {
                        'delta': {
                            'partial_json': function_info.get('arguments'),
                            'type': 'input_json_delta'
                        },
                        'index': content_index,
                        'type': 'content_block_delta'
                    }

                    # 尝试解析当前的 JSON 状态
                    try:
                        current_json = json.loads(tool_calls_data[tool_index]['arguments'])
                    except:
                        current_json = {}

                    yield {
                        'type': 'input_json',
                        'partial_json': function_info.get('arguments'),
                        'snapshot': current_json
                    }

        # 处理结束
        if finish_reason == 'stop':
            if accumulated_content:
                yield {
                    'index': 0,
                    'type': 'content_block_stop',
                    'content_block': {
                        'citations': None,
                        'text': accumulated_content,
                        'type': 'text'
                    }
                }
            content_finished = True
        elif finish_reason == 'tool_calls':
            # 为每个工具调用发送 content_block_stop
            for idx, tool_data in tool_calls_data.items():
                try:
                    parsed_args = json.loads(tool_data['arguments']) if tool_data['arguments'] else {}
                except:
                    parsed_args = tool_data['arguments']

                content_index = idx + (1 if accumulated_content else 0)
                yield {
                    'index': content_index,
                    'type': 'content_block_stop',
                    'content_block': {
                        'id': tool_data['id'],
                        'input': parsed_args,
                        'name': tool_data['name'],
                        'type': 'tool_use'
                    }
                }

async def convert_input_anthropic_to_openai_format(request_body: Dict) -> Dict:
    """
    将Anthropic风格的输入参数转换为OpenAI风格的输入参数
    """
    # 1. messages转换
    messages = []
    for msg in request_body.get("messages", []):
        role = msg.get("role")
        content = msg.get("content")
        # content 可能是字符串或content block数组
        if isinstance(content, str):
            messages.append({"role": role, "content": content})
        elif isinstance(content, list):
            ctnt = ""
            for block in content:
                block_type = block.get("type")
                if block_type == "text":
                    text_content = block.get("text", "")
                    ctnt += f"\n{text_content}"
                elif block_type == "tool_use":
                    # ctnt += f"\n[Use Tool: {block.get('name')} with input {block.get('input')}]"
                    ctnt += f"\n{block}]"
                elif block_type == "tool_result":
                    # ctnt += f"\n[Tool Result: {block.get('name')} output {block.get('content')}]"
                    ctnt += f"\n{block}]"
                else:
                    # ctnt += f"\n{block}]"  # 直接转换为字符串
                    raise NotImplementedError(f"Block type {block_type} not implemented in conversion.")
            messages.append({"role": role, "content": ctnt})
        else:
            messages.append({"role": role, "content": ""})

    # 2. model
    model = request_body.get("model")
    # 3. max_tokens
    max_tokens = request_body.get("max_tokens")
    # 4. temperature
    temperature = request_body.get("temperature")
    # 5. top_p
    top_p = request_body.get("top_p")
    # 6. stop_sequences -> stop
    stop = request_body.get("stop_sequences")
    # 7. metadata
    metadata = request_body.get("metadata")
    # 8. service_tier
    service_tier = request_body.get("service_tier")
    # 9. stream
    stream = request_body.get("stream")
    # 10. tools
    tools = request_body.get("tools")
    # 很复杂的转换：
    oai_tools = []
    if tools:
        for i, tool in enumerate(tools):
            input_schema = tool.get("input_schema", {})
            oai_tool = {
                "type": "function",
                "function": {
                    "name": tool.get("name"),
                    "description": tool.get("description"),
                    "parameters": input_schema,
                    # "strict": True,
                }
            }
            oai_tools.append(oai_tool)

    # 11. tool_choice
    tool_choice = request_body.get("tool_choice")
    # 12. system prompt
    system = request_body.get("system")
    if system:
        # 多条system合并为一条
        system_list = [x["text"] for x in system if isinstance(x, dict) and "text" in x]
        # _system_text = "\n\n".join(system_list)

        # OpenAI的system prompt是messages的第一条
        # messages = [{"role": "system", "content": _system_text}] + messages
        for sys_text in reversed(system_list):
            messages.insert(0, {"role": "system", "content": sys_text})


    # 构造OpenAI格式
    openai_body = {
        "model": model,
        "messages": messages,
    }
    if max_tokens is not None:
        openai_body["max_completion_tokens"] = max_tokens
    if temperature is not None:
        openai_body["temperature"] = temperature
    if top_p is not None:
        openai_body["top_p"] = top_p
    if stop is not None:
        openai_body["stop"] = stop
    if metadata is not None:
        openai_body["metadata"] = metadata
    if service_tier is not None:
        openai_body["service_tier"] = service_tier
    if stream is not None:
        openai_body["stream"] = stream
    if oai_tools:
        openai_body["tools"] = oai_tools
    if tool_choice is not None:
        openai_body["tool_choice"] = tool_choice

    return openai_body


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ...._models import BaseModel

__all__ = ["RefusalContentBlock"]


class RefusalContentBlock(BaseModel):
    refusal: str

    type: Literal["refusal"]
    """Always `refusal`."""





from typing import List, Dict, Union
from ...base_class._worker_class import HRemoteModel
from starlette.applications import Starlette
from starlette.routing import Mount
# import contextlib
from contextlib import AsyncExitStack


# @contextlib.asynccontextmanager
# async def lifespan(app: Starlette):
#     async with contextlib.AsyncExitStack() as stack:
#         # await stack.enter_async_context(echo_mcp.session_manager.run())
#         await stack.enter_async_context(mcp.session_manager.run())
#         yield
        
        
def build_lifespan_for_starlette(models: List[HRemoteModel]):
    """
    创建适用于 Starlette 的 lifespan 函数
    """
    # 筛选出启用了 MCP 的模型
    models_with_mcp = [model for model in models if model.config.enable_mcp]
    
    async def _lifespan(app: Starlette):
        async with AsyncExitStack() as stack:
            # 为所有启用了 MCP 的模型启动会话
            for model in models_with_mcp:
                await stack.enter_async_context(model.mcp.session_manager.run())
            yield
    
    return _lifespan if models_with_mcp else None

def build_mcp_kwargs_for_starlette(
    models: HRemoteModel | List[HRemoteModel],
    route_prefix: str = "/apiv2"
    ):
    """
    To adapte to mcp server, the additional kwargs are needed.
    referes to: https://github.com/modelcontextprotocol/python-sdk?tab=readme-ov-file
    """
    
    if not isinstance(models, list):
        models = [models]
        
    # 筛选出启用了 MCP 的模型
    models_with_mcp = [model for model in models if model.config.enable_mcp]
    if not models_with_mcp:  # no models with MCP enabled
        return {}
    
    mcp_transports = set(model.config.mcp_transport for model in models_with_mcp)
    if len(mcp_transports) > 1:
        raise ValueError(f"All models must have the same mcp_transport, but got: {mcp_transports}")
    mcp_transport = mcp_transports.pop()
    
    routes = []
    # lifespan = None
        
    for model in models:
        if not model.config.enable_mcp:
            continue
        # routes.append(Mount("/math", model.mcp.streamable_http_app()))
        # route_path = f'{route_prefix}/{model.name}'
        route_path = f"{route_prefix}/mcp/{model.model_id}"  # full path: /apiv2/mcp/{model_id}
        
        if mcp_transport == "streamable-http":
            routes.append(Mount(route_path, model.mcp.streamable_http_app()))
        elif mcp_transport == "sse":
            routes.append(Mount(route_path, model.mcp.sse_app()))
        else:
            raise ValueError(f"Unsupported mcp_transport: {mcp_transport}, only 'sse' and 'streamable-http' are supported.")
    
    
    
    if mcp_transport == "streamable-http":
        # 创建 lifespan 函数
        lifespan_func = build_lifespan_for_starlette(models_with_mcp)
        return {
            "routes": routes,
            "lifespan": lifespan_func
        }
    else:  # sse does not need lifespan
        return {
            "routes": routes,
            # "lifespan": lifespan
        }
            


"""
HepAI对MCP的封装，实现将MCP工具方便地搭载到Worker节点上，同时支持MCP协议和HepAI无限函数协议。
"""

from typing import Literal
from mcp.server.fastmcp import FastMCP
# from hepai import HRModel, HModelConfig, HWorkerConfig, HWorkerAPP

from ...base_class._worker_class import HRModel, HModelConfig
from ...worker._worker_class import HWorkerConfig
from ...worker.worker_app import HWorkerAPP


class HaiMCP(FastMCP):
    """HepAI对MCP的封装，实现将MCP工具方便地搭载到Worker节点上，同时支持MCP协议和HepAI无限函数协议。"""

    def __init__(self, **kwargs):
        """初始化HaiMCP实例，继承自FastMCP。"""
        super().__init__(**kwargs)
        
        self.mcp_config_dict = kwargs  # 保存初始化参数以备后续使用
        
        
    def run(
        self,
        transport: Literal["stdio", "sse", "streamable-http"] = "stdio",
        mount_path: str | None = None,
    ) -> None:
        
        super().run(transport=transport, mount_path=mount_path)
        
        
    def run_via_worker(
        self,
        transport: Literal["stdio", "sse", "streamable-http"] = "stdio",
        mount_path: str | None = None,
        model_config: HModelConfig | None = None,
        worker_config: HWorkerConfig | None = None,
    ) -> None:
        """把MCP搭载到WorkerModel上"""
        
        model_cfg = model_config if model_config is not None else HModelConfig()
        model_cfg.enable_mcp = True
        model_cfg.mcp_transport = transport
        
        if "/" in str(self.name):
            name = self.name
        else:
            name = f'hepai/{self.name}'  # 自动添加前缀
            
        model = HRModel(
            name=name,
            config=model_cfg,
            mcp=self,  # 把自己传给model
        )
        
        config_dict = self.mcp_config_dict.copy()
        config_dict = {
            k: v for k, v in config_dict.items() if k in HWorkerConfig.__dataclass_fields__
        }
        # worker_cfg = HWorkerConfig.from_dict(config_dict)
        worker_cfg = worker_config if worker_config is not None else HWorkerConfig()
        worker_cfg.update_from_dict(config_dict)
        
        app = HWorkerAPP(
            models=[model],
            worker_config=worker_cfg,
        )
        print(app.worker.get_worker_info(), flush=True)
        
        import uvicorn
        uvicorn.run(app, host=app.host, port=app.port)
        
        


"""
damei nn api for ai algorithm management (unified ai interface)
"""

import argparse
from ..version import __backend__, __version__, __author__, __appname__, __email__
from ..version import __affiliation__, __version_suffix__
import os, sys
from pathlib import Path

# damei.nn warpper
__pydir__ = Path(os.path.dirname(os.path.abspath(__file__)))

# 导入damei lib
try:
    import damei
except:
    sys.path.append(f'{__pydir__.parent.parent.parent}/damei')
    import damei

logger = damei.getLogger('hai')
# logger.info(f'HAI version: {__version__}')


if __backend__ == 'local':
    from .basic.base import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
    from .basic.registry import MODULES, SCRIPTS, IOS, InitRegister
    from .basic.utils import Config
    from .basic.grpc import grpc_secure_server
    from ..uaii.uaii_main import UAII
    from ..uaii.cli.cli_main import CommandLineInterface
    from ..testor import Testor
    from ..uaii.datasets.datasets_hub import DatasetsHub
    from .basic import argparse, parse_args_into_dataclasses, parse_args
    from ..uaii.utils import general
    from ..uaii.worker.worker import WorkerWarper as worker, WorkerArgs
    from ..uaii.worker.base_worker_model import BaseWorkerModel

elif __backend__ == 'damei':
    from damei.nn.api.base import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
    from damei.nn.api.registry import MODULES, SCRIPTS, IOS
    from damei.nn.api.utils import Config
    from damei.nn.api.utils import Config as PyConfigLoader
    from damei.nn.api import UAII as UAII
else:
    raise NotImplementedError(f'{__backend__} backend is not supported, only local and damei, please check')

from .workers_api.llm.llm import HaiLLM as LLM
from .workers_api.model import HaiModel as Model
from .workers_api.model import HaiModel as Models


# hai_config = Config(f'{__pydir__.parent}/configs/Base/hai_config.py')
# from ..configs.Base.hai_config import root_path, weights_root
# root_path = f'{Path(__pydir__).parent.parent}'

from ..configs import CONST

# init_register = InitRegister(internal_dir=hai_config.root_path)
init_register = InitRegister(internal_dir=CONST.ROOT_PATH)
uaii = UAII()
cli = CommandLineInterface(
    uaii=uaii, 
    api_fold_name=CONST.API_FOLD_NAME,
    root_path=CONST.ROOT_PATH,
    )
api_key = os.getenv('HEPAI_API_KEY')



from urllib.request import Request, urlopen
from urllib.parse import urlencode
import ssl

data = {'ip': '192.168.30.101', 'port': 8848, 'serviceName': 'ai-service', 'weight': 1.0, 'enable': True, 'healthy': True, 'clusterName': None, 'ephemeral': True, 'groupName': 'DEFAULT_GROUP', 'namespaceId': 'nacos.naming.serviceName'}
# data = None
server_url = 'http://192.169.30.101:8848'
url = '/nacos/v1/ns/instance?username=nacos&password=nacos'
url = '/nacos/v1/ns/instance'
url = "/nacos/v1/ns/instance?port=8848&healthy=true&ip=11.11.11.11&weight=1.0&serviceName=ai-service&encoding=GBK&namespaceId=n1'"
all_headers = {}
method = 'POST'



req = Request(url=server_url + url, data=urlencode(data).encode() if data else None,
              headers=all_headers, method=method)
# req = Request(url=se)
get_instances_list = 'http://192.168.30.101:8848/nacos/v1/ns/instance/list?serviceName=ai-service'
get_service_list = 'http://192.168.30.101:8848/nacos/v1/ns/service/list?pageNo=1&pageSize=10'
add_service = 'http://192.168.30.101:8848/nacos/v1/ns/service?serviceName=ai-service&metadata=k1%3dv1'
add_instance = f'http://192.168.30.101:8848/nacos/v1/ns/instance?port=8848&healthy=true&ip=11.11.11.11' \
               f'&weight=1.0&serviceName=ai-service&encoding=GBK&namespaceId=n1'


def do(_url, method='GET'):
    req = Request(url=_url, method=method, headers=all_headers,
                  data=urlencode(data).encode() if data else None)
    ctx = ssl.SSLContext()
    timeout = 3
    resp = urlopen(req, timeout=timeout, context=ctx)
    print(resp.read().decode())

print('查询实例列表')
do(get_instances_list)
do(get_service_list)
try:
    do(add_service, method='POST')
    do(get_service_list)
except:
    pass

print('注册实例')
# add_instance = f'http://192.168.30.101:8848/nacos/v1/ns/instance?port=50052&healthy=true&ip=192.168.30.99' \
#                f'&weight=1.0&serviceName=ai-service&encoding=GBK&namespaceId=n1'
add_instance = f'http://192.168.30.101:8848/nacos/v1/ns/instance?port=50052&ip=192.168.30.99&serviceName=ai-service' \
               # f''
do(add_instance, method='POST')
print('查询实例列表')
do(get_instances_list)
print('查询实例详情')
get_instance_detail = f'http://192.168.30.101:8848/nacos/v1/ns/instance?' \
                      f'serviceName=ai-service&ip=192.168.30.99&port=50052&cluster=DEFAULT'
do(get_instance_detail)



from typing import List, Callable, Dict
from dataclasses import field, dataclass
from fastapi import APIRouter, Depends, HTTPException
from fastapi.requests import Request
from fastapi import FastAPI

# from ..auth.authorizer import HAuthorizer, HAPIKeyAuth
# from ...configs.model_aliases import model_aliases_to_real_name
# from ...utils.http_parsing_util import read_request_body
# from ...utils._logger import Logger, save_minitor_log
# from ...utils import general
from .._worker_class import CommonWorker
from ..utils import(
    read_request_body
    )
from ..worker_app import HWorkerAPP, FunctionParamsItem
from ..singletons import authorizer

# logger = Logger.get_logger("llm_router_group")
# api_key_auth: Callable = Depends(HAuthorizer().api_key_auth)
api_key_auth: Callable = Depends(authorizer.api_key_auth)


@dataclass
class LLMRouterGroup:
    name: str = "llm"
    prefix: str = "/apiv2"
    tags: List[str] = field(default_factory=lambda: ["llm"])
    router: APIRouter = field(default_factory=APIRouter)
    parent_app: HWorkerAPP = None  # type: ignore

    def __post_init__(self):
        # super().__post_init__()
        
        self.count = 0
        
        rt = self.router
        rt.post("/completions", dependencies=[api_key_auth])(self.chat_completions)
        rt.post("/v1/completions", dependencies=[api_key_auth])(self.chat_completions)
        rt.post("/chat/completions", dependencies=[api_key_auth])(self.chat_completions)
        rt.post("/v1/chat/completions", dependencies=[api_key_auth])(self.chat_completions)
        rt.post("/v1/embeddings", dependencies=[api_key_auth])(self.embeddings)
        rt.post("/embeddings", dependencies=[api_key_auth])(self.embeddings)
        rt.get("/models", dependencies=[api_key_auth])(self.list_models)
        rt.get("/v1/models", dependencies=[api_key_auth])(self.list_models)
        rt.post("/images/generations", dependencies=[api_key_auth])(self.image_generations)
        rt.post("/v1/images/generations", dependencies=[api_key_auth])(self.image_generations)

        # rerank route
        rt.post("/rerank", dependencies=[api_key_auth])(self.rerank)
        rt.post("/v1/rerank", dependencies=[api_key_auth])(self.rerank)

        # openai
        rt.post("/responses", dependencies=[api_key_auth])(self.responses)
        rt.post("/v1/responses", dependencies=[api_key_auth])(self.responses)


    async def chat_completions(self, request: Request):
        request_body: Dict = await read_request_body(request=request)
        if "model" not in request_body:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] This `model` must be specified")
        model = request_body["model"]
        self.count += 1
        # await save_minitor_log(logger, user_auth)
        func_params = FunctionParamsItem(
            args=[],
            kwargs=request_body
        )
        rst = await self.parent_app.worker_unified_gate(
            function_params=func_params,
            model=model, 
            function="chat_completions",
        )
        return rst

    async def embeddings(self, request: Request, user_auth = api_key_auth):
        request_body: Dict = await read_request_body(request=request)
        if "model" not in request_body:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] This `model` must be specified")
        model = request_body["model"]
        self.count += 1
        # if user_auth.resc_attr.resource_type == "worker":
        #     request_body = self.update_request_body_for_worker(request_body, user_auth)
        # await save_minitor_log(logger, user_auth)
        
        func_params = FunctionParamsItem(
            args=[],
            kwargs=request_body
        )
        rst = await self.parent_app.worker_unified_gate(
            function_params=func_params,
            model=model, 
            function="embeddings",
        )
        return rst
        # return await self.worker.unified_gate_async(
        #     model=model, 
        #     function="embeddings",
        #     args=[],
        #     kwargs=request_body,
        # )
        
    async def rerank(self, request: Request, user_auth = api_key_auth):
        request_body: Dict = await read_request_body(request=request)
        if "model" not in request_body:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] This `model` must be specified")
        model = request_body["model"]
        self.count += 1
        # if user_auth.resc_attr.resource_type == "worker":
        #     request_body = self.update_request_body_for_worker(request_body, user_auth)
        # await save_minitor_log(logger, user_auth)
        
        func_params = FunctionParamsItem(
            args=[],
            kwargs=request_body
        )
        rst = await self.parent_app.worker_unified_gate(
            function_params=func_params,
            model=model, 
            function="rerank",
        )
        return rst

    async def list_models(self, user_auth = api_key_auth):
        return await self.parent_app.get_models()

    async def image_generations(self, request: Request, user_auth = api_key_auth):
        request_body: Dict = await read_request_body(request=request)
        if "prompt" not in request_body:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] The 'prompt' field is required")
        if "model" not in request_body:
            request_body["model"] = "dall-e-2"
        model = request_body["model"]
        valid_models = ["dall-e-2", "dall-e-3", "gpt-image-1"]
        model_without_prefix = model.split("/")[-1]
        if model_without_prefix not in valid_models:
            raise HTTPException(status_code=400, detail=f"[LLMRouterGroup] Invalid model '{model}'. Must be one of {valid_models}")
        prompt = request_body["prompt"]
        max_lengths = {"gpt-image-1": 32000, "dall-e-2": 1000, "dall-e-3": 4000}
        if len(prompt) > max_lengths[model_without_prefix]:
            raise HTTPException(status_code=400, detail=f"[LLMRouterGroup] Prompt too long for {model}. Maximum length is {max_lengths[model]} characters")
        request_body.setdefault("n", 1)
        request_body.setdefault("size", "auto" if model_without_prefix == "gpt-image-1" else "1024x1024")
        request_body.setdefault("quality", "auto")
        request_body.setdefault("response_format", "url")
        if model_without_prefix == "dall-e-3" and request_body["n"] != 1:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] For dall-e-3, only n=1 is supported")
        if request_body["n"] < 1 or request_body["n"] > 10:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] n must be between 1 and 10")
        if user_auth.resc_attr.resource_type == "worker":
            request_body = self.update_request_body_for_worker(request_body, user_auth)
        self.count += 1
        await save_minitor_log(logger, user_auth)
        return await self.worker.unified_gate_async(
            model=model, 
            function="image_generations",
            args=[],
            kwargs=request_body,
        )

    def update_request_body_for_worker(self, request_body: Dict, user_auth) -> Dict:
        resc = user_auth.resc_attr.resource
        if resc.type == "litellm":
            request_body["api_key"] = user_auth.api_key
            request_channel = user_auth.resc_attr.request_channel.name
            request_body["extra_body"] = {
                "user": user_auth.user_attr.username,
                "metadata": {"tags": [f"channelTag: {request_channel}"]}
            }
        elif resc.type == "drsai":
            u = user_auth.user_attr
            user_info_dict = {
                "id": u.id,
                "username": u.username,
                "user_level": u.user_level.to_str(),
                "user_groups": [g.name for g in u.user_groups],
                "email": u.email,
                "balance": u.balance,
                "api_key": user_auth.api_key
            }
            request_body["extra_body"] = {
                "user": user_info_dict,
                "base_models": "deepseek/deepseek-v3",
            }
        return request_body

    async def responses(self, request: Request, user_auth=api_key_auth):
        request_body: Dict = await read_request_body(request=request)
        if "model" not in request_body:
            raise HTTPException(status_code=400, detail="[LLMRouterGroup] This `model` must be specified")

        model = request_body["model"]
        self.count += 1

        func_params = FunctionParamsItem(
            args=[],
            kwargs=request_body
        )

        rst = await self.parent_app.worker_unified_gate(
            function_params=func_params,
            model=model,
            function="responses",
        )
        return rst
    
    


"""
HepAI - Custom Remote Model
"""
import json
try:
    from hepai import __version__
except:
    import os, sys
    from pathlib import Path
    here = Path(__file__).parent
    sys.path.insert(1, str(here.parent.parent))
    from hepai import __version__

from hepai import HRModel, HWorkerAPP, HWorkerConfig  # Import the HRModel class from the hepai package.

from mcp.server.fastmcp import FastMCP

# mcp = FastMCP(name="MathServer", stateless_http=True)

class SimpleWorkerModel(HRModel):  # Define a custom worker model inheriting from HRModel.
    def __init__(self, name: str = "hepai/simple-model", **kwargs):
        super().__init__(name=name, **kwargs)

    # @HRModel.mcp.tool()
    @HRModel.remote_callable   # Decorate the function to enable remote call.
    async def simple_method(self, a: int = 1, b: int = 2) -> int:
        """Define your custom method here."""
        return a + b
    
    
import contextlib
from starlette.applications import Starlette
from starlette.routing import Mount
# Create a combined lifespan to manage both session managers

model = SimpleWorkerModel(name="hepai/custom-model")  # Instantiate the custom worker model.
mcp = model.mcp

@contextlib.asynccontextmanager
async def lifespan(app: Starlette):
    async with contextlib.AsyncExitStack() as stack:
        # await stack.enter_async_context(echo_mcp.session_manager.run())
        await stack.enter_async_context(mcp.session_manager.run())
        yield

if __name__ == "__main__":
    # SimpleWorkerModel.run(
    #     routes=[
    #         Mount("/", math_mcp.streamable_http_app()),
    #     ]
    #     )  # Run the custom worker model.


    import uvicorn
    app = HWorkerAPP(
        model, worker_config=HWorkerConfig(),
        routes=[
            Mount("/math", mcp.streamable_http_app()),
        ], 
        lifespan=lifespan
        )  # Instantiate the APP, which is a FastAPI application.
    print(app.worker.get_worker_info(), flush=True)
    # 启动服务
    uvicorn.run(app, host=app.host, port=app.port)